from enum import Enum
from ..colors import FMT_PARSE, FMT_NONE
from .privmodes import PrivModes
from .csr_constants import MCAUSE_TRANSLATION
from .exceptions import RiscemuBaseException
from .int32 import UInt32
from .instruction import InstructionWithEncoding


class CpuTrapType(Enum):
    TIMER = 1
    SOFTWARE = 2
    EXTERNAL = 3
    EXCEPTION = 4


class CpuTrap(BaseException):
    code: int
    """
    31-bit value encoding the exception code in the mstatus register
    """
    interrupt: int
    """
    The isInterrupt bit in the mstatus register
    """

    mtval: UInt32
    """
    contents of the mtval register
    """

    type: CpuTrapType
    """
    The type (timer, external, software) of the trap
    """

    priv: PrivModes
    """
    The privilege level this trap targets
    """

    def __init__(
        self, code: int, mtval, type: CpuTrapType, priv: PrivModes = PrivModes.MACHINE
    ):
        self.interrupt = 0 if type == CpuTrapType.EXCEPTION else 1
        self.code = code
        self.mtval = UInt32(mtval)
        self.priv = priv
        self.type = type

    @property
    def mcause(self):
        return (self.interrupt << 31) + self.code

    def message(self) -> str:
        return ""

    def __repr__(self):
        name = "Reserved interrupt({}, {})".format(self.interrupt, self.code)

        if (self.interrupt, self.code) in MCAUSE_TRANSLATION:
            name = MCAUSE_TRANSLATION[(self.interrupt, self.code)] + "({}, {})".format(
                self.interrupt, self.code
            )

        return "{} {{priv={}, type={}, mtval={:x}}} {}".format(
            name, self.priv.name, self.type.name, self.mtval, self.message()
        )

    def __str__(self):
        return self.__repr__()


class IllegalInstructionTrap(CpuTrap):
    def __init__(self, ins: InstructionWithEncoding):
        super().__init__(2, ins.encoding, CpuTrapType.EXCEPTION)


class InstructionAddressMisalignedTrap(CpuTrap):
    def __init__(self, addr: int):
        super().__init__(0, addr, CpuTrapType.EXCEPTION)


class InstructionAccessFault(CpuTrap):
    def __init__(self, addr: int):
        super().__init__(1, addr, CpuTrapType.EXCEPTION)


class TimerInterrupt(CpuTrap):
    def __init__(self):
        super().__init__(7, 0, CpuTrapType.TIMER)


class EcallTrap(CpuTrap):
    def __init__(self, mode: PrivModes):
        super().__init__(mode.value + 8, 0, CpuTrapType.EXCEPTION)


class InvalidElfException(RiscemuBaseException):
    def __init__(self, msg: str):
        super().__init__()
        self.msg = msg

    def message(self):
        return (
            FMT_PARSE + '{}("{}")'.format(self.__class__.__name__, self.msg) + FMT_NONE
        )


class LoadAccessFault(CpuTrap):
    def __init__(self, msg, addr, size, op):
        super(LoadAccessFault, self).__init__(5, addr, CpuTrapType.EXCEPTION)
        self.msg = msg
        self.addr = addr
        self.size = size
        self.op = op

    def message(self):
        return "(During {} at 0x{:08x} of size {}: {})".format(
            self.op, self.addr, self.size, self.msg
        )



"""
RiscEmu (c) 2021 Anton Lydike

SPDX-License-Identifier: MIT

This package aims at providing an all-round usable RISC-V emulator and debugger

It contains everything needed to run assembly files, so you don't need any custom compilers or toolchains
"""

# to read package version:
import importlib.metadata

__author__ = "Anton Lydike <Anton@Lydike.com>"
__copyright__ = "Copyright 2023 Anton Lydike"
__version__ = importlib.metadata.version(__name__)



"""
RiscEmu (c) 2021 Anton Lydike

SPDX-License-Identifier: MIT
"""

from .instruction_set import *


class RV32M(InstructionSet):
    """
    The RV32M Instruction set, containing multiplication and division instructions
    """

    def instruction_mul(self, ins: "Instruction"):
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set(rd, rs1 * rs2)

    def instruction_mulh(self, ins: "Instruction"):
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set(rd, Int32((rs1.signed().value * rs2.signed().value) >> 32))

    def instruction_mulhsu(self, ins: "Instruction"):
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set(rd, Int32((rs1.signed().value * rs2.unsigned_value) >> 32))

    def instruction_mulhu(self, ins: "Instruction"):
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set(rd, UInt32((rs1.unsigned_value * rs2.unsigned_value) >> 32))

    def instruction_div(self, ins: "Instruction"):
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set(rd, rs1 // rs2)

    def instruction_divu(self, ins: "Instruction"):
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins, signed=False)
        self.regs.set(rd, rs1 // rs2)

    def instruction_rem(self, ins: "Instruction"):
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set(rd, rs1 % rs2)

    def instruction_remu(self, ins: "Instruction"):
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins, signed=False)
        self.regs.set(rd, rs1 % rs2)



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Optional

from ..colors import FMT_MEM, FMT_NONE, FMT_UNDERLINE, FMT_ORANGE, FMT_ERROR
from ..helpers import format_bytes
from . import (
    MemoryFlags,
    T_AbsoluteAddress,
    InstructionContext,
    T_RelativeAddress,
    Instruction,
    Int32,
)


@dataclass
class MemorySection(ABC):
    name: str
    flags: MemoryFlags
    size: int
    base: T_AbsoluteAddress
    owner: str
    context: InstructionContext

    @property
    def end(self):
        return self.base + self.size

    @abstractmethod
    def read(self, offset: T_RelativeAddress, size: int) -> bytearray:
        pass

    @abstractmethod
    def write(self, offset: T_RelativeAddress, size: int, data: bytearray):
        pass

    @abstractmethod
    def read_ins(self, offset: T_RelativeAddress) -> Instruction:
        pass

    def dump(
        self,
        start: T_RelativeAddress,
        end: Optional[T_RelativeAddress] = None,
        fmt: Optional[str] = None,
        bytes_per_row: Optional[int] = None,
        rows: int = 10,
        group: Optional[int] = None,
        highlight: Optional[int] = None,
    ):
        """
        Dump the section. If no end is given, the rows around start are printed and start is highlighted.

        :param start: The address to start at
        :param end: The end of the printed space
        :param fmt: either hex, int, char or asm
        :param bytes_per_row: the number of bytes displayed per row
        :param rows: the number of rows displayed
        :param group: Group this many bytes into one when displaying
        :param highlight: Highlight the group containing this address
        :return:
        """
        if isinstance(start, Int32):
            start = start.value
        if isinstance(end, Int32):
            end = end.value

        if fmt is None:
            if self.flags.executable and self.flags.read_only:
                bytes_per_row = 4
                fmt = "asm"
            else:
                fmt = "hex"

        if fmt == "char":
            if bytes_per_row is None:
                bytes_per_row = 4
            if group is None:
                group = 1

        if group is None:
            group = 4

        if bytes_per_row is None:
            bytes_per_row = 4

        if fmt not in ("asm", "hex", "int", "char"):
            print(
                FMT_ERROR
                + "[MemorySection] Unknown format {}, known formats are {}".format(
                    fmt, ", ".join(("asm", "hex", "int", "char"))
                )
                + FMT_NONE
            )

        if end is None:
            end = min(start + (bytes_per_row * (rows // 2)), self.size)
            highlight = start
            start = max(0, start - (bytes_per_row * (rows // 2)))

        if fmt == "asm":
            print(
                FMT_MEM
                + "{}, viewing {} instructions:".format(self, (end - start) // 4)
                + FMT_NONE
            )

            for addr in range(start, end, 4):
                if addr == highlight:
                    print(FMT_UNDERLINE + FMT_ORANGE, end="")
                print(
                    "0x{:04x}: {}{}".format(
                        self.base + addr, self.read_ins(addr), FMT_NONE
                    )
                )
        else:
            print(
                FMT_MEM + "{}, viewing {} bytes:".format(self, (end - start)) + FMT_NONE
            )

            aligned_end = (
                end - (end % bytes_per_row) if end % bytes_per_row != 0 else end
            )

            for addr in range(start, aligned_end, bytes_per_row):
                hi_ind = (highlight - addr) // group if highlight is not None else -1
                print(
                    "0x{:04x}: {}{}".format(
                        self.base + addr,
                        format_bytes(
                            self.read(addr, bytes_per_row), fmt, group, hi_ind
                        ),
                        FMT_NONE,
                    )
                )

            if aligned_end != end:
                hi_ind = (
                    (highlight - aligned_end) // group if highlight is not None else -1
                )
                print(
                    "0x{:04x}: {}{}".format(
                        self.base + aligned_end,
                        format_bytes(
                            self.read(aligned_end, end % bytes_per_row),
                            fmt,
                            group,
                            hi_ind,
                        ),
                        FMT_NONE,
                    )
                )

    def dump_all(
        self,
        fmt: Optional[str] = None,
        bytes_per_row: Optional[int] = None,
        rows: int = 10,
        group: Optional[int] = None,
        highlight: Optional[int] = None,
    ):
        self.dump(0, self.size, fmt, bytes_per_row, rows, group, highlight)

    def __repr__(self):
        return "{}[{}] at 0x{:08X} (size={}bytes, flags={}, owner={})".format(
            self.__class__.__name__,
            self.name,
            self.base,
            self.size,
            self.flags,
            self.owner,
        )

    def debug_str(self) -> str:
        return "0x{:08X}-0x{:08X} flags={} {}[{}] ({}bytes)".format(
            self.base,
            self.base + self.size,
            self.flags,
            self.owner,
            self.name,
            self.size,
        )



"""
RiscEmu (c) 2021-2022 Anton Lydike

SPDX-License-Identifier: MIT
"""

from dataclasses import dataclass


@dataclass(frozen=True, init=True)
class RunConfig:
    stack_size: int = 8 * 1024 * 64  # for 8KB stack
    include_scall_symbols: bool = True
    add_accept_imm: bool = False
    # debugging
    debug_instruction: bool = True
    debug_on_exception: bool = True
    # allowed syscalls
    scall_input: bool = True
    scall_fs: bool = False
    verbosity: int = 0
    slowdown: float = 1
    unlimited_registers: bool = False
    flen: int = 64
    # runtime config
    use_libc: bool = False
    ignore_exit_code: bool = False
    # csr stuff:
    # frequency of the real-time clock
    rtclock_tickrate: int = 32768



import math

from typing import Any, Union, ClassVar
from ctypes import c_int32, c_uint32


class Int32:
    """
    This class implements 32bit signed integers (see :class:`UInt32` for unsigned integers)

    It implements basically all mathematical dunder magic methods (__add__, __sub__, etc.)

    You can use it just like you would any other integer, just be careful when passing it
    to functions which actually expect an integer and not an Int32.
    """

    _type = c_int32
    __slots__ = ("_val",)

    MIN_VALUE: ClassVar[int] = -(2**31)
    MAX_VALUE: ClassVar[int] = 2**31 - 1

    def __init__(
        self, val: Union[int, c_int32, c_uint32, "Int32", bytes, bytearray, bool] = 0
    ):
        if isinstance(val, (c_uint32, c_int32, Int32)):
            self._val = self.__class__._type(val.value)
        elif isinstance(val, (int, bool)):
            self._val = self.__class__._type(val)
        elif isinstance(val, (bytes, bytearray)):
            signed = len(val) == 4 and self._type == c_int32
            self._val = self.__class__._type(
                int.from_bytes(val, "little", signed=signed)
            )
        else:
            raise RuntimeError(
                "Unknown {} input type: {} ({})".format(
                    self.__class__.__name__, type(val), val
                )
            )

    def __add__(self, other: Union["Int32", int]):
        if isinstance(other, Int32):
            other = other.value

        return self.__class__(self._val.value + other)

    def __sub__(self, other: Union["Int32", int]):
        if isinstance(other, Int32):
            other = other.value
        return self.__class__(self._val.value - other)

    def __mul__(self, other: Union["Int32", int]):
        if isinstance(other, Int32):
            other = other.value
        return self.__class__(self._val.value * other)

    def __truediv__(self, other: Any):
        return self // other

    def __floordiv__(self, other: Any):
        if isinstance(other, Int32):
            other = other.value
        return self.__class__(self.value // other)

    def __mod__(self, other: Union["Int32", int]):
        if isinstance(other, Int32):
            other = other.value
        return self.__class__(self._val.value % other)

    def __and__(self, other: Union["Int32", int]):
        if isinstance(other, Int32):
            other = other.value
        return self.__class__(self._val.value & other)

    def __or__(self, other: Union["Int32", int]):
        if isinstance(other, Int32):
            other = other.value
        return self.__class__(self._val.value | other)

    def __xor__(self, other: Union["Int32", int]):
        if isinstance(other, Int32):
            other = other.value
        return self.__class__(self._val.value ^ other)

    def __lshift__(self, other: Union["Int32", int]):
        if isinstance(other, Int32):
            other = other.value
        return self.__class__(self.value << other)

    def __rshift__(self, other: Union["Int32", int]):
        if isinstance(other, Int32):
            other = other.value
        return self.__class__(self.value >> other)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, int):
            return self.value == other
        elif isinstance(other, Int32):
            return self.value == other.value
        return False

    def __neg__(self):
        return self.__class__(-self._val.value)

    def __invert__(self):
        return self.__class__(~self.value)

    def __abs__(self):
        return self.__class__(abs(self.value))

    def __bytes__(self):
        return self.to_bytes(4)

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.value)

    def __str__(self):
        return str(self.value)

    def __format__(self, format_spec: str):
        return self.value.__format__(format_spec)

    def __hash__(self):
        return hash(self.value)

    def __gt__(self, other: Any):
        if isinstance(other, Int32):
            other = other.value
        return self.value > other

    def __lt__(self, other: Any):
        if isinstance(other, Int32):
            other = other.value
        return self.value < other

    def __le__(self, other: Any):
        if isinstance(other, Int32):
            other = other.value
        return self.value <= other

    def __ge__(self, other: Any):
        if isinstance(other, Int32):
            other = other.value
        return self.value >= other

    def __bool__(self):
        return bool(self.value)

    def __cmp__(self, other: Any):
        if isinstance(other, Int32):
            other = other.value
        return self.value.__cmp__(other)

    # right handed binary operators

    def __radd__(self, other: Any):
        return self + other

    def __rsub__(self, other: Any):
        return self.__class__(other) - self

    def __rmul__(self, other: Any):
        return self * other

    def __rtruediv__(self, other: Any):
        return self.__class__(other) // self

    def __rfloordiv__(self, other: Any):
        return self.__class__(other) // self

    def __rmod__(self, other: Any):
        return self.__class__(other) % self

    def __rand__(self, other: Any):
        return self.__class__(other) & self

    def __ror__(self, other: Any):
        return self.__class__(other) | self

    def __rxor__(self, other: Any):
        return self.__class__(other) ^ self

    @property
    def value(self) -> int:
        """
        The value represented by this Integer
        :return:
        """
        return self._val.value

    def unsigned(self) -> "UInt32":
        """
        Convert to an unsigned representation. See :class:Uint32
        :return:
        """
        return UInt32(self)

    def to_bytes(self, bytes: int = 4) -> bytearray:
        """
        Convert to a bytearray of length :param:bytes

        :param bytes: The length of the bytearray
        :return: A little-endian representation of the contained integer
        """
        return bytearray(self.unsigned_value.to_bytes(4, "little"))[0:bytes]

    def signed(self) -> "Int32":
        """
        Convert to a signed representation. See :class:Int32
        :return:
        """
        return self

    @property
    def unsigned_value(self):
        """
        Return the value interpreted as an unsigned integer
        :return:
        """
        return c_uint32(self.value).value

    def shift_right_logical(self, amount: Union["Int32", int]) -> "Int32":
        """
        This function implements logical right shifts, meaning that the sign bit is shifted as well.

        This is equivalent to (self.value % 0x100000000) >> amount

        :param amount: Number of positions to shift
        :return: A new Int32 object representing the shifted value (keeps the signed-ness of the source)
        """
        if isinstance(amount, Int32):
            amount = amount.value
        return self.__class__((self.value % 0x100000000) >> amount)

    def __int__(self):
        return self.value

    def __hex__(self):
        return hex(self.value)

    @classmethod
    def sign_extend(cls, data: Union[bytes, bytearray, int], bits: int):
        """
        Create an instance of Int32 by sign extending :param:bits bits from :param:data
        to 32 bits

        :param data: The source data
        :param bits: The number of bits in the source data
        :return: An instance of Int32, holding the sign-extended value
        """
        if isinstance(data, (bytes, bytearray)):
            data = int.from_bytes(data, "little")
        sign = data >> (bits - 1)
        if sign > 1:
            print("overflow in Int32.sext!")
        if sign:
            data = (data & (2 ** (bits - 1) - 1)) - 2 ** (bits - 1)
        return cls(data)

    @classmethod
    def from_float(cls, number: float) -> "Int32":
        """
        Convert a floating point number to an instance of Int32, handling overflows and NaN according to RISC-V spec.

        - Valid values and infinities are clamped between MIN_VALUE and MAX_VALUE
        - NaN is treated as -inf
        """
        if math.isnan(number):
            number = cls.MIN_VALUE
        elif number > cls.MAX_VALUE:
            number = cls.MAX_VALUE
        elif number < cls.MIN_VALUE:
            number = cls.MIN_VALUE
        else:
            number = int(number)
        return cls(number)


class UInt32(Int32):
    """
    An unsigned version of :class:Int32.
    """

    _type = c_uint32

    MIN_VALUE: ClassVar[int] = 0
    MAX_VALUE: ClassVar[int] = 2**32 - 1

    def unsigned(self) -> "UInt32":
        """
        Return a new instance representing the same bytes, but signed
        :return:
        """
        return self

    @property
    def unsigned_value(self) -> int:
        return self._val.value

    def shift_right_logical(self, amount: Union["Int32", int]) -> "UInt32":
        """
        see :meth:`Int32.shift_right_logical <Int32.shift_right_logical>`

        :param amount: Number of positions to shift
        :return: A new Int32 object representing the shifted value (keeps the signed-ness of the source)
        """
        if isinstance(amount, Int32):
            amount = amount.value
        return UInt32(self.value >> amount)

    def signed(self) -> "Int32":
        # Overflow is handled natively by the underlying c_uint32 to c_int32 conversion!
        return Int32(self._val)



from abc import ABC
from typing import Optional

from riscemu.core import MemorySection, MemoryFlags, T_RelativeAddress


class IOModule(MemorySection, ABC):
    def __init__(
        self,
        name: str,
        flags: MemoryFlags,
        size: int,
        owner: str = "system",
        base: int = 0,
    ):
        super(IOModule, self).__init__(name, flags, size, base, owner, None)

    def contains(self, addr, size: int = 0):
        return (
            self.base <= addr < self.base + self.size
            and self.base <= addr + size <= self.base + self.size
        )

    def dump(self, *args, **kwargs):
        print(self)

    def __repr__(self):
        return "{}[{}] at 0x{:0X} (size={}bytes, flags={})".format(
            self.__class__.__name__, self.name, self.base, self.size, self.flags
        )



from typing import ClassVar, Generic, TypeVar, Tuple, Iterable, Callable, Type

from .instruction_set import InstructionSet, Instruction
from riscemu.core import BaseFloat, CPU, INS_NOT_IMPLEMENTED, UInt32

_FloatT = TypeVar("_FloatT", bound=BaseFloat)


class FloatArithBase(Generic[_FloatT], InstructionSet):
    flen: ClassVar[int]
    _float_cls: ClassVar[Type[BaseFloat]]

    def __init__(self, cpu: CPU):
        assert cpu.regs.flen >= self.flen, "{} implies cpu flen of at least {}".format(
            self.__class__.__name__, self.flen
        )
        super().__init__(cpu)

    def base_fmadd(self, ins: Instruction):
        """
        :Format:
        | fmadd.{s,d,q}    rd,rs1,rs2,rs3

        :Description:
        | Perform fused multiply addition.

        :Implementation:
        | f[rd] = f[rs1]×f[rs2]+f[rs3]

        """
        rd, rs1, rs2, rs3 = self.parse_rd_rs_rs_rs(ins)
        self.regs.set_f(rd, rs1 * rs2 + rs3)

    def base_fmsub(self, ins: Instruction):
        """
        :Format:
        | fmsub.{s,d,q}    rd,rs1,rs2,rs3

        :Description:
        | Perform fused multiply subtraction.

        :Implementation:
        | f[rd] = f[rs1]×f[rs2]-f[rs3]
        """
        rd, rs1, rs2, rs3 = self.parse_rd_rs_rs_rs(ins)
        self.regs.set_f(rd, rs1 * rs2 - rs3)

    def base_fnmsub(self, ins: Instruction):
        """
        :Format:
        | fnmsub.{s,d,q}   rd,rs1,rs2,rs3

        :Description:
        | Perform fused negated multiply addition.

        :Implementation:
        | f[rd] = -f[rs1]×f[rs2]+f[rs3]
        """
        rd, rs1, rs2, rs3 = self.parse_rd_rs_rs_rs(ins)
        self.regs.set_f(rd, -rs1 * rs2 + rs3)

    def base_fnmadd(self, ins: Instruction):
        """
        :Format:
        | fnmadd.{s,d,q}   rd,rs1,rs2,rs3

        :Description:
        | Perform single-precision fused negated multiply addition.

        :Implementation:
        | f[rd] = -f[rs1]×f[rs2]-f[rs3]
        """
        rd, rs1, rs2, rs3 = self.parse_rd_rs_rs_rs(ins)
        self.regs.set_f(rd, -rs1 * rs2 - rs3)

    def base_fadd(self, ins: Instruction):
        """
        +-----+-----+-----+-----+-----+-----+-----+---+
        |31-27|26-25|24-20|19-15|14-12|11-7 |6-2  |1-0|
        +-----+-----+-----+-----+-----+-----+-----+---+
        |00000|00   |rs2  |rs1  |rm   |rd   |10100|11 |
        +-----+-----+-----+-----+-----+-----+-----+---+


        :Format:
        | fadd.s     rd,rs1,rs2

        :Description:
        | Perform single-precision floating-point addition.

        :Implementation:
        | f[rd] = f[rs1] + f[rs2]
        """
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set_f(
            rd,
            rs1 + rs2,
        )

    def base_fsub(self, ins: Instruction):
        """
        :Format:
        | fsub.{s,d,q}     rd,rs1,rs2

        :Description:
        | Perform single-precision floating-point subtraction.

        :Implementation:
        | f[rd] = f[rs1] - f[rs2]
        """
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set_f(
            rd,
            rs1 - rs2,
        )

    def base_fmul(self, ins: Instruction):
        """
        :Format:
        | fmul.{s,d,q}     rd,rs1,rs2

        :Description:
        | Perform floating-point multiplication.

        :Implementation:
        | f[rd] = f[rs1] × f[rs2]
        """
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set_f(
            rd,
            rs1 * rs2,
        )

    def base_fdiv(self, ins: Instruction):
        """
        :Format:
        | fdiv.{s,d,q}     rd,rs1,rs2

        :Description:
        | Perform floating-point division.

        :Implementation:
        | f[rd] = f[rs1] / f[rs2]
        """
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set_f(
            rd,
            rs1 / rs2,
        )

    def base_fsqrt(self, ins: Instruction):
        """
        :Format:
        | fsqrt.{s,d,q}    rd,rs1

        :Description:
        | Perform single-precision square root.

        :Implementation:
        | f[rd] = sqrt(f[rs1])
        """
        rd, rs = self.parse_rd_rs(ins)
        self.regs.set_f(rd, self._float_cls.bitcast(self.regs.get_f(rs)) ** 0.5)

    def base_fsgnj(self, ins: Instruction):
        """
        :Format:
        | fsgnj.{s,d,q}    rd,rs1,rs2

        :Description:
        | Produce a result that takes all bits except the sign bit from rs1.
        | The result’s sign bit is rs2’s sign bit.

        :Implementation:
        | f[rd] = {f[rs2][31], f[rs1][30:0]}
        """
        INS_NOT_IMPLEMENTED(ins)

    def base_fsgnjn(self, ins: Instruction):
        """
        :Format:
        | fsgnjn.{s,d,q}   rd,rs1,rs2

        :Description:
        | Produce a result that takes all bits except the sign bit from rs1.
        | The result’s sign bit is opposite of rs2’s sign bit.

        :Implementation:
        | f[rd] = {~f[rs2][31], f[rs1][30:0]}
        """
        INS_NOT_IMPLEMENTED(ins)

    def base_fsgnjx(self, ins: Instruction):
        """
        :Format:
        | fsgnjx.{s,d,q}   rd,rs1,rs2

        :Description:
        | Produce a result that takes all bits except the sign bit from rs1.
        | The result’s sign bit is XOR of sign bit of rs1 and rs2.

        :Implementation:
        | f[rd] = {f[rs1][31] ^ f[rs2][31], f[rs1][30:0]}
        """
        INS_NOT_IMPLEMENTED(ins)

    def base_fmin(self, ins: Instruction):
        """
        :Format:
        | fmin.{s,d,q}     rd,rs1,rs2

        :Description:
        | Write the smaller of rs1 and rs2 to rd.

        :Implementation:
        | f[rd] = min(f[rs1], f[rs2])
        """
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set_f(
            rd,
            min(
                rs1,
                rs2,
            ),
        )

    def base_fmax(self, ins: Instruction):
        """
        :Format:
        | fmax.{s,d,q}     rd,rs1,rs2

        :Description:
        | Write the larger of rs1 and rs2 to rd.

        :Implementation:
        | f[rd] = max(f[rs1], f[rs2])
        """
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set_f(
            rd,
            max(
                rs1,
                rs2,
            ),
        )

    def base_feq(self, ins: Instruction):
        """
        :Format:
        | feq.{s,d,q}      rd,rs1,rs2

        :Description:
        | Performs a quiet equal comparison between floating-point registers rs1 and rs2 and record the Boolean result in integer register rd.
        | Only signaling NaN inputs cause an Invalid Operation exception.
        | The result is 0 if either operand is NaN.

        :Implementation:
        | x[rd] = f[rs1] == f[rs2]
        """
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set(rd, UInt32(rs1 == rs2))

    def base_flt(self, ins: Instruction):
        """
        :Format:
        | flt.{s,d,q}      rd,rs1,rs2

        :Description:
        | Performs a quiet less comparison between floating-point registers rs1 and rs2 and record the Boolean result in integer register rd.
        | Only signaling NaN inputs cause an Invalid Operation exception.
        | The result is 0 if either operand is NaN.

        :Implementation:
        | x[rd] = f[rs1] < f[rs2]
        """
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set(rd, UInt32(rs1 < rs2))

    def base_fle(self, ins: Instruction):
        """
        :Format:
        | fle.{s,d,q}      rd,rs1,rs2

        :Description:
        | Performs a quiet less or equal comparison between floating-point registers rs1 and rs2 and record the Boolean result in integer register rd.
        | Only signaling NaN inputs cause an Invalid Operation exception.
        | The result is 0 if either operand is NaN.

        :Implementation:
        | x[rd] = f[rs1] <= f[rs2]
        """
        rd, rs1, rs2 = self.parse_rd_rs_rs(ins)
        self.regs.set(rd, UInt32(rs1 <= rs2))

    def base_fclass(self, ins: Instruction):
        """
        :Format:
        | fclass.{s,d,q}   rd,rs1

        :Description:
        | Examines the value in floating-point register rs1 and writes to integer register rd a 10-bit mask that indicates the class of the floating-point number.
        | The format of the mask is described in [classify table]_.
        | The corresponding bit in rd will be set if the property is true and clear otherwise.
        | All other bits in rd are cleared. Note that exactly one bit in rd will be set.

        :Implementation:
        | x[rd] = classifys(f[rs1])
        """
        INS_NOT_IMPLEMENTED(ins)

    def base_load(self, ins: Instruction):
        """
        :Format:
          | fl{w,d,q}        rd,offset(rs1)

        :Description:
          | Load a floating-point value from memory into floating-point register rd.

        :Implementation:
          | f[rd] = M[x[rs1] + sext(offset)][31:0]
        """
        rd, addr = self.parse_mem_ins(ins)
        bytes = self.mmu.read(addr.value, self.flen // 8)
        self.regs.set_f(rd, self._float_cls(bytes))

    def base_save(self, ins: Instruction):
        """:Format:
          | fs{w,d,q}        rs2,offset(rs1)

        :Description:
          | Store a value from floating-point register rs2 to memory.

        :Implementation:
          | M[x[rs1] + sext(offset)] = f[rs2][31:0]
        """
        rs, addr = self.parse_mem_ins(ins)
        val = self._float_cls.bitcast(self.regs.get_f(rs))
        self.mmu.write(addr.value, self.flen // 8, bytearray(val.bytes))

    def get_instructions(self) -> Iterable[Tuple[str, Callable[[Instruction], None]]]:
        yield from super().get_instructions()

        qual = {32: "s", 64: "d", 128: "q"}.get(self.flen)
        load_save_qual = {32: "w", 64: "d", 128: "q"}.get(self.flen)

        yield from (
            ("fmadd." + qual, self.base_fmadd),
            ("fmsub." + qual, self.base_fmsub),
            ("fnmsub." + qual, self.base_fnmsub),
            ("fnmadd." + qual, self.base_fnmadd),
            ("fadd." + qual, self.base_fadd),
            ("fsub." + qual, self.base_fnmadd),
            ("fmul." + qual, self.base_fmul),
            ("fdiv." + qual, self.base_fdiv),
            ("fsqrt." + qual, self.base_fsqrt),
            ("fsgnj." + qual, self.base_fsgnj),
            ("fsgnjn." + qual, self.base_fsgnjn),
            ("fsgnjx." + qual, self.base_fsgnjx),
            ("fmin." + qual, self.base_fmin),
            ("fmax." + qual, self.base_fmax),
            ("feq." + qual, self.base_feq),
            ("flt." + qual, self.base_flt),
            ("fle." + qual, self.base_fle),
            ("fl" + load_save_qual, self.base_load),
            ("fs" + load_save_qual, self.base_save),
        )

    def parse_rd_rs(self, ins: Instruction) -> Tuple[str, str]:
        assert len(ins.args) == 2
        return ins.get_reg(0), ins.get_reg(1)

    def parse_rd_rs_rs(self, ins: Instruction) -> Tuple[str, _FloatT, _FloatT]:
        assert len(ins.args) == 3
        return (
            ins.get_reg(0),
            self._float_cls.bitcast(self.regs.get_f(ins.get_reg(1))),
            self._float_cls.bitcast(self.regs.get_f(ins.get_reg(2))),
        )

    def parse_rd_rs_rs_rs(
        self, ins: Instruction
    ) -> Tuple[str, _FloatT, _FloatT, _FloatT]:
        assert len(ins.args) == 4
        return (
            ins.get_reg(0),
            self._float_cls.bitcast(self.regs.get_f(ins.get_reg(1))),
            self._float_cls.bitcast(self.regs.get_f(ins.get_reg(2))),
            self._float_cls.bitcast(self.regs.get_f(ins.get_reg(3))),
        )



"""
RiscEmu (c) 2021 Anton Lydike

SPDX-License-Identifier: MIT

This package holds all instruction sets, available to the processor
"""

from .instruction_set import InstructionSet, Instruction
from .RV32M import RV32M
from .RV32I import RV32I
from .RV32A import RV32A
from .RV32F import RV32F
from .RV32D import RV32D
from .RV_Debug import RV_Debug
from .Zicsr import Zicsr

InstructionSetDict = {
    v.__name__: v for v in [RV32I, RV32M, RV32A, RV32F, RV32D, Zicsr, RV_Debug]
}

__all__ = [
    "Instruction",
    "InstructionSet",
    "InstructionSetDict",
    "RV32I",
    "RV32M",
    "RV32A",
    "RV32F",
    "RV32D",
    "Zicsr",
    "RV_Debug",
]



"""
RiscEmu (c) 2021-2022 Anton Lydike

SPDX-License-Identifier: MIT

This file contains the CPU logic (not the individual instruction sets). See instructions/instruction_set.py for more info
on them.
"""
import typing
from typing import List, Type

from ..config import RunConfig
from ..colors import FMT_CPU, FMT_NONE, FMT_ERROR, FMT_GRAY, FMT_CYAN
from ..debug import launch_debug_session
from ..syscall import SyscallInterface, get_syscall_symbols
from . import (
    CPU,
    Int32,
    BinaryDataMemorySection,
    MMU,
    RiscemuBaseException,
    LaunchDebuggerException,
    PrivModes,
    Instruction,
    SimpleInstruction,
)

if typing.TYPE_CHECKING:
    from ..instructions import InstructionSet


class UserModeCPU(CPU):
    """
    This class represents a single CPU. It holds references to it's mmu, registers and syscall interrupt handler.

    It is initialized with a configuration and a list of instruction sets.
    """

    def __init__(self, instruction_sets: List[Type["InstructionSet"]], conf: RunConfig):
        """
        Creates a CPU instance.

        :param instruction_sets: A list of instruction set classes. These must inherit from the InstructionSet class
        """
        # setup CPU states
        super().__init__(MMU(), instruction_sets, conf)

        self.exit_code = 0

        # setup syscall interface
        self.syscall_int = SyscallInterface()

        # add global syscall symbols, but don't overwrite any user-defined symbols
        syscall_symbols = get_syscall_symbols()
        syscall_symbols.update(self.mmu.global_symbols)
        self.mmu.global_symbols.update(syscall_symbols)
        self.mode = PrivModes.USER

    def step(self, verbose: bool = False):
        """
        Execute a single instruction, then return.
        """
        launch_debugger = False

        try:
            self.cycle += 1
            ins = self.mmu.read_ins(self.pc)
            if verbose:
                if self.conf.verbosity > 2:
                    ins_str = self._format_ins(ins)
                else:
                    ins_str = str(ins)
                print(FMT_CPU + "   0x{:08X}:{} {}".format(self.pc, FMT_NONE, ins_str))
            self.pc += self.INS_XLEN
            self.run_instruction(ins)
        except RiscemuBaseException as ex:
            if isinstance(ex, LaunchDebuggerException):
                # if the debugger is active, raise the exception to
                if self.debugger_active:
                    raise ex

                print(FMT_CPU + "[CPU] Debugger launch requested!" + FMT_NONE)
                launch_debugger = True
            else:
                print(ex.message())
                ex.print_stacktrace()
                print(FMT_CPU + "[CPU] Halting due to exception!" + FMT_NONE)
                self.halted = True

        if launch_debugger:
            launch_debug_session(self)

    def run(self, verbose: bool = False):
        while not self.halted:
            self.step(verbose)

        if self.conf.verbosity > 0:
            print(
                FMT_CPU
                + "[CPU] Program exited with code {}".format(self.exit_code)
                + FMT_NONE
            )

    def setup_stack(self, stack_size: int = 1024 * 4) -> bool:
        """
        Create program stack and populate stack pointer
        :param stack_size: the size of the required stack, defaults to 4Kib
        :return:
        """
        stack_sec = BinaryDataMemorySection(
            bytearray(stack_size),
            ".stack",
            None,  # FIXME: why does a binary data memory section require a context?
            "",
            0,
        )

        if not self.mmu.load_section(stack_sec, fixed_position=False):
            print(FMT_ERROR + "[CPU] Could not insert stack section!" + FMT_NONE)
            return False

        self.regs.set("sp", Int32(stack_sec.base + stack_sec.size))

        if self.conf.verbosity > 1:
            print(
                FMT_CPU
                + "[CPU] Created stack of size {} at 0x{:x}".format(
                    stack_size, stack_sec.base
                )
                + FMT_NONE
            )

        return True

    def _format_arg(self, arg: str, ins: Instruction) -> str:
        if arg in self.regs.vals or arg in self.regs.valid_regs:
            return "{}{}=0x{:x}{}".format(
                arg, FMT_GRAY, self.regs.get(arg, False), FMT_NONE
            )
        elif arg in self.regs.float_vals or arg in self.regs.float_regs:
            return "{}{}={}{}".format(arg, FMT_GRAY, self.regs.get_f(arg), FMT_NONE)
        elif isinstance(ins, SimpleInstruction):
            val = ins.context.resolve_label(arg)
            if val is None:
                val = ins.context.resolve_numerical_label(arg, ins.addr)
            if val is None:
                return FMT_CYAN + arg + FMT_NONE
            return "{}{}=0x{:x}{}".format(arg, FMT_GRAY, val, FMT_NONE)

    def _format_ins(self, ins: Instruction) -> str:
        args = ", ".join(self._format_arg(arg, ins) for arg in ins.args)
        return "{}\t{}".format(ins.name, args)



from .formats import (
    INSTRUCTION_ARGS_DECODER,
    op,
    decode_i,
    decode_r,
    decode_u,
    decode_b,
    decode_j,
    decode_s,
    decode_i_shamt,
    decode_i_unsigned,
)
from .regs import RISCV_REGS


def int_to_hex(num: int):
    if num < 0:
        return f"-0x{-num:x}"
    return f"0x{num:x}"


def format_ins(ins: int, name: str, fmt: str = "int"):
    opcode = op(ins)
    if fmt == "hex":
        fmt = int_to_hex
    else:
        fmt = str

    if opcode not in INSTRUCTION_ARGS_DECODER:
        return f"{name} <unknown op>"

    decoder = INSTRUCTION_ARGS_DECODER[opcode]
    if name in ("ecall", "ebreak", "mret", "sret", "uret"):
        return name
    if opcode in (0x8, 0x0):
        r1, r2, imm = decoder(ins)
        return f"{name:<7} {RISCV_REGS[r1]}, {imm}({RISCV_REGS[r2]})"
    elif decoder in (decode_i, decode_i_unsigned, decode_b, decode_i_shamt, decode_s):
        r1, r2, imm = decoder(ins)
        r1, r2 = RISCV_REGS[r1], RISCV_REGS[r2]
        return f"{name:<7} {r1}, {r2}, {fmt(imm)}"
    elif decoder in (decode_r,):
        rd, rs1, rs2 = [RISCV_REGS[x] for x in decoder(ins)]
        return f"{name:<7} {rd}, {rs1}, {rs2}"
    elif decoder in (decode_j, decode_u):
        r1, imm = decoder(ins)
        return f"{name:<7} {RISCV_REGS[r1]}, {fmt(imm)}"

    return f"{name} <unknown decoder>"



"""
RiscEmu (c) 2023 Anton Lydike

SPDX-License-Identifier: MIT
"""

from collections import defaultdict
from typing import Type

from ..helpers import *

from . import Int32, BaseFloat


class Registers:
    """
    Represents a bunch of registers
    """

    valid_regs = {
        "zero",
        "ra",
        "sp",
        "gp",
        "tp",
        "s0",
        "fp",
        "t0",
        "t1",
        "t2",
        "t3",
        "t4",
        "t5",
        "t6",
        "s1",
        "s2",
        "s3",
        "s4",
        "s5",
        "s6",
        "s7",
        "s8",
        "s9",
        "s10",
        "s11",
        "a0",
        "a1",
        "a2",
        "a3",
        "a4",
        "a5",
        "a6",
        "a7",
    }

    float_regs = {
        "ft0",
        "ft1",
        "ft2",
        "ft3",
        "ft4",
        "ft5",
        "ft6",
        "ft7",
        "ft8",
        "ft9",
        "ft10",
        "ft11",
        "fs0",
        "fs1",
        "fs2",
        "fs3",
        "fs4",
        "fs5",
        "fs6",
        "fs7",
        "fs8",
        "fs9",
        "fs10",
        "fs11",
        "fa0",
        "fa1",
        "fa2",
        "fa3",
        "fa4",
        "fa5",
        "fa6",
        "fa7",
    }

    flen: int
    _float_type: Type[BaseFloat]

    def __init__(self, infinite_regs: bool = False, flen: int = 32):
        self.vals: dict[str, Int32] = defaultdict(UInt32)
        self.flen = flen
        self._float_type = BaseFloat.flen_to_cls(flen)
        self.float_vals: dict[str, BaseFloat] = defaultdict(self._float_type)

        self.last_set = None
        self.last_read = None

        self.infinite_regs = infinite_regs

    def dump(self, full: bool = False):
        """
        Dump all registers to stdout
        :param full: If True, floating point registers are dumped too
        """
        named_regs = [self._reg_repr(reg) for reg in Registers.named_registers()]

        lines: list[list[str]] = [[] for _ in range(12)]
        if not full:
            regs = [("a", 8), ("s", 12), ("t", 7)]
        else:
            regs = [
                ("a", 8),
                ("s", 12),
                ("t", 7),
                ("ft", 8),
                ("fa", 8),
                ("fs", 12),
            ]
        for name, s in regs:
            for i in range(12):
                if i >= s:
                    lines[i].append(" " * 15)
                else:
                    reg = "{}{}".format(name, i)
                    lines[i].append(self._reg_repr(reg))

        print(
            "Registers[{},{}](".format(
                FMT_ORANGE + FMT_UNDERLINE + "read" + FMT_NONE,
                FMT_ORANGE + FMT_BOLD + "written" + FMT_NONE,
            )
        )
        if not full:
            print("\t" + " ".join(named_regs[0:3]))
            print("\t" + " ".join(named_regs[3:]))
            print("\t" + "--------------- " * 3)
        else:
            print("\t" + " ".join(named_regs))
            print("\t" + "--------------- " * 6)
        for line in lines:
            print("\t" + " ".join(line))
        print(")")

    def dump_reg_a(self):
        """
        Dump the a registers
        """
        print(
            "Registers[a]:"
            + " ".join(self._reg_repr("a{}".format(i)) for i in range(8))
        )

    def _reg_repr(self, reg: str, name_len=4, fmt="08X"):
        if reg in self.float_regs:
            txt = "{:{}}={: .3e}".format(reg, name_len, self.get_f(reg, False))
        else:
            txt = "{:{}}=0x{:{}}".format(reg, name_len, self.get(reg, False), fmt)
        if reg == "fp":
            reg = "s0"
        if reg == self.last_set:
            return FMT_ORANGE + FMT_BOLD + txt + FMT_NONE
        if reg == self.last_read:
            return FMT_ORANGE + FMT_UNDERLINE + txt + FMT_NONE
        if reg == "zero":
            return txt
        should_grayscale_int = (
            reg in self.valid_regs
            and self.get(reg, False) == 0
            and reg not in Registers.named_registers()
        )
        should_grayscale_float = reg in self.float_regs and self.get_f(reg, False) == 0
        if should_grayscale_int or should_grayscale_float:
            return FMT_GRAY + txt + FMT_NONE
        return txt

    def set(self, reg: str, val: Int32, mark_set: bool = True) -> bool:
        """
        Set a register content to val
        :param reg: The register to set
        :param val: The new value
        :param mark_set: If True, marks this register as "last accessed" (only used internally)
        :return: If the operation was successful
        """

        if reg == "zero":
            return False
        # if reg not in Registers.all_registers():
        #    raise InvalidRegisterException(reg)
        # replace fp register with s1, as these are the same register
        if reg == "fp":
            reg = "s1"
        if mark_set:
            self.last_set = reg

        if not self.infinite_regs and reg not in self.valid_regs:
            raise RuntimeError("Invalid register: {}".format(reg))

        self.vals[reg] = val.signed()
        return True

    def get(self, reg: str, mark_read: bool = True) -> Int32:
        """
        Returns the contents of register reg
        :param reg: The register name
        :param mark_read: If the register should be marked as "last read" (only used internally)
        :return: The contents of register reg
        """
        # if reg not in Registers.all_registers():
        #    raise InvalidRegisterException(reg)
        if reg == "fp":
            reg = "s0"

        if not self.infinite_regs and reg not in self.valid_regs:
            raise RuntimeError("Invalid register: {}".format(reg))

        if mark_read:
            self.last_read = reg
        return self.vals[reg]

    def get_f(self, reg: str) -> BaseFloat:
        if not self.infinite_regs and reg not in self.float_regs:
            raise RuntimeError("Invalid float register: {}".format(reg))
        return self.float_vals[reg]

    def set_f(self, reg: str, val: BaseFloat):
        if not self.infinite_regs and reg not in self.float_regs:
            raise RuntimeError("Invalid float register: {}".format(reg))
        self.float_vals[reg] = self._float_type.bitcast(val)

    @staticmethod
    def named_registers():
        """
        Return all named registers
        :return: The list
        """
        return ["zero", "ra", "sp", "gp", "tp", "fp"]

    def __repr__(self):
        return "<Registers[xlen=32,flen={}]{}>".format(
            self.flen,
            "{"
            + ", ".join(self._reg_repr("a{}".format(i), 2, "0x") for i in range(8))
            + "}",
        )



from .instruction_table import *
from typing import Tuple, List


def print_ins(ins: int):
    print("  f7      rs2   rs1   f3  rd    op")
    print(
        f"0b{ins >> 25 :07b}_{(ins >> 20) & 0b11111:05b}_{(ins >> 15) & 0b11111:05b}_{(ins >> 12) & 0b111:03b}_{(ins >> 7) & 0b11111:05b}_{ins & 0b1111111:07b}"
    )


STATIC_INSN: Dict[int, Tuple[str, List[int], int]] = {
    0x00000013: ("nop", [], 0x00000013),
    0x00008067: ("ret", [], 0x00008067),
    0xFE010113: ("addi", [2, 2, -32], 0xFE010113),
    0x02010113: ("addi", [2, 2, 32], 0x02010113),
    0x00100073: ("ebreak", [], 0x00100073),
    0x00000073: ("ecall", [], 0x00000073),
    0x30200073: ("mret", [], 0x30200073),
    0x00200073: ("uret", [], 0x00200073),
    0x10200073: ("sret", [], 0x10200073),
    0x10500073: ("wfi", [], 0x10500073),
}


def int_from_ins(insn: bytearray):
    return int.from_bytes(insn, "little")


def name_from_insn(ins: int):
    opcode = op(ins)
    if opcode not in RV32:
        print_ins(ins)
        raise RuntimeError(f"Invalid opcode: {opcode:0x} in insn {ins:x}")
    dec = RV32[opcode]

    if isinstance(dec, str):
        return dec

    fun3 = funct3(ins)
    if fun3 not in dec:
        print_ins(ins)
        raise RuntimeError(f"Invalid funct3: {fun3:0x} in insn {ins:x}")

    dec = dec[fun3]
    if isinstance(dec, str):
        return dec

    if opcode == 0x1C and fun3 == 0:
        # we have ecall/ebreak
        token = imm110(ins)
        if token in dec:
            return dec[token]
        print_ins(ins)
        raise RuntimeError(f"Invalid instruction in ebreak/ecall region: {ins:x}")

    fun7 = funct7(ins)
    if opcode == 0b1011 and fun3 == 0b10:
        # ignore the two aq/lr bits located in the fun7 block
        # riscemu has no memory reordering, therefore we don't need to look at these bits ever
        fun7 = fun7 >> 2

    if fun7 in dec:
        if opcode == 0x0C or (opcode == 0x04 and fun3 == 5):
            dec = dec[fun7]
            return dec
        print("unknown instruction?!")
        return dec[fun7]

    print_ins(ins)
    raise RuntimeError(f"Invalid instruction: {ins:x}")


def decode(ins: Union[bytearray, bytes]) -> Tuple[str, List[int], int]:
    insn = int_from_ins(ins)

    if insn & 3 != 3:
        print_ins(insn)
        raise RuntimeError("Not a RV32 instruction!")

    if insn in STATIC_INSN:
        return STATIC_INSN[insn]

    opcode = op(insn)
    if opcode not in INSTRUCTION_ARGS_DECODER:
        print_ins(insn)
        raise RuntimeError("No instruction decoder found for instruction")

    return name_from_insn(insn), INSTRUCTION_ARGS_DECODER[opcode](insn), insn



from .IOModule import IOModule
from core.traps import InstructionAccessFault
from ..core import T_RelativeAddress, Instruction, MemoryFlags, Int32


class TextIO(IOModule):
    def read_ins(self, offset: T_RelativeAddress) -> Instruction:
        raise InstructionAccessFault(self.base + offset)

    def __init__(self, base: int, buflen: int = 128):
        super(TextIO, self).__init__(
            "TextIO", MemoryFlags(False, False), buflen + 4, base=base
        )
        self.buff = bytearray(buflen)
        self.current_line = ""

    def read(self, addr: int, size: int) -> bytearray:
        raise InstructionAccessFault(self.base + addr)

    def write(self, addr: int, size: int, data: bytearray):
        if addr == 0:
            if size > 4:
                raise InstructionAccessFault(addr)
            if Int32(data) != 0:
                self._print()
                return
        buff_start = addr - 4
        self.buff[buff_start : buff_start + size] = data[0:size]

    def _print(self):
        buff = self.buff
        self.buff = bytearray(self.size)
        if b"\x00" in buff:
            buff = buff.split(b"\x00")[0]
        text = buff.decode("ascii")
        if "\n" in text:
            lines = text.split("\n")
            lines[0] = self.current_line + lines[0]
            for line in lines[:-1]:
                self._present(line)
            self.current_line = lines[-1]
        else:
            self.current_line += text

    def _present(self, text: str):
        print("[TextIO:{:x}] {}".format(self.base, text))



"""
RiscEmu (c) 2023 Anton Lydike

SPDX-License-Identifier: MIT

This file contains copious amounts of docstrings that were all taken
from https://msyksphinz-self.github.io/riscv-isadoc/html/rvfd.html
(all the docstrings on the instruction methods documenting the opcodes
and their function)
"""
from typing import Tuple

from .instruction_set import InstructionSet, Instruction
from .float_base import FloatArithBase
from riscemu.core import INS_NOT_IMPLEMENTED, Float32, Int32, UInt32, Float64


class RV32D(FloatArithBase[Float64]):
    flen = 64
    _float_cls = Float64

    def instruction_fcvt_d_w(self, ins: Instruction):
        """
        +-----+-----+-----+-----+-----+-----+-----+---+
        |31-27|26-25|24-20|19-15|14-12|11-7 |6-2  |1-0|
        +-----+-----+-----+-----+-----+-----+-----+---+
        |11010|01   |00000|rs1  |rm   |rd   |10100|11 |
        +-----+-----+-----+-----+-----+-----+-----+---+

        :Format:
          | fcvt.d.w   rd,rs1

        :Description:
          | Converts a 32-bit signed integer, in integer register rs1 into a double-precision floating-point number in floating-point register rd.

        :Implementation:
          | x[rd] = sext(s32_{f64}(f[rs1]))
        """
        rd, rs = self.parse_rd_rs(ins)
        self.regs.set_f(rd, Float64(self.regs.get(rs).value))

    def instruction_fcvt_d_wu(self, ins: Instruction):
        """
        +-----+-----+-----+-----+-----+-----+-----+---+
        |31-27|26-25|24-20|19-15|14-12|11-7 |6-2  |1-0|
        +-----+-----+-----+-----+-----+-----+-----+---+
        |11010|01   |00001|rs1  |rm   |rd   |10100|11 |
        +-----+-----+-----+-----+-----+-----+-----+---+

        :Format:
          | fcvt.d.wu  rd,rs1

        :Description:
          | Converts a 32-bit unsigned integer, in integer register rs1 into a double-precision floating-point number in floating-point register rd.

        :Implementation:
          | f[rd] = f64_{u32}(x[rs1])
        """
        rd, rs = self.parse_rd_rs(ins)
        self.regs.set_f(rd, Float64(self.regs.get(rs).unsigned_value))

    def instruction_fcvt_w_d(self, ins: Instruction):
        """
        +-----+-----+-----+-----+-----+-----+-----+---+
        |31-27|26-25|24-20|19-15|14-12|11-7 |6-2  |1-0|
        +-----+-----+-----+-----+-----+-----+-----+---+
        |11000|01   |00000|rs1  |rm   |rd   |10100|11 |
        +-----+-----+-----+-----+-----+-----+-----+---+

        :Format:
          | fcvt.w.d   rd,rs1

        :Description:
          | Converts a double-precision floating-point number in floating-point register rs1 to a signed 32-bit integer, in integer register rd.

        :Implementation:
          | x[rd] = sext(s32_{f64}(f[rs1]))
        """
        rd, rs = self.parse_rd_rs(ins)
        self.regs.set(rd, Int32.from_float(self.regs.get_f(rs).value))

    def instruction_fcvt_wu_d(self, ins: Instruction):
        """
        +-----+-----+-----+-----+-----+-----+-----+---+
        |31-27|26-25|24-20|19-15|14-12|11-7 |6-2  |1-0|
        +-----+-----+-----+-----+-----+-----+-----+---+
        |11000|01   |00001|rs1  |rm   |rd   |10100|11 |
        +-----+-----+-----+-----+-----+-----+-----+---+

        :Format:
          | fcvt.wu.d  rd,rs1

        :Description:
          | Converts a double-precision floating-point number in floating-point register rs1 to a unsigned 32-bit integer, in integer register rd.

        :Implementation:
          | x[rd] = sext(u32f64(f[rs1]))
        """
        rd, rs = self.parse_rd_rs(ins)
        self.regs.set(rd, UInt32.from_float(self.regs.get_f(rs).value))



from .cpu import CPU
from .memory_section import MemorySection

from typing import List


class Platform:
    """
    This is a wrapper around a given hardware configuration.
    """

    harts: List[CPU]

    memory: List[MemorySection]

    def __init__(self):
        pass

    def step(self):
        performed_step = False

        for cpu in self.harts:
            if not cpu.halted:
                cpu.step(cpu.conf.verbosity > 1)
                performed_step = True

        return performed_step



from typing import List, Optional, Set

from ..colors import FMT_RED, FMT_BOLD, FMT_NONE, FMT_MEM
from ..helpers import get_section_base_name
from . import InstructionContext, T_AbsoluteAddress, MemorySection


class Program:
    """
    This represents a collection of sections which together form an executable program

    When you want to create a program which can be located anywhere in memory, set base to None,
    this signals the other components, that this is relocatable. Set the base of each section to
    the offset in the program, and everything will be taken care of for you.

    """

    name: str
    context: InstructionContext
    global_labels: Set[str]
    relative_labels: Set[str]
    sections: List[MemorySection]
    base: Optional[T_AbsoluteAddress]
    is_loaded: bool

    @property
    def size(self):
        if len(self.sections) == 0:
            return 0
        if self.base is None:
            return self.sections[-1].base + self.sections[-1].size
        return (self.sections[-1].base - self.base) + self.sections[-1].size

    def __init__(self, name: str, base: Optional[int] = None):
        self.name = name
        self.context = InstructionContext()
        self.sections = []
        self.global_labels = set()
        self.relative_labels = set()
        self.base = base
        self.is_loaded = False

    def add_section(self, sec: MemorySection):
        # print a warning when a section is located before the programs base
        if self.base is not None:
            if sec.base < self.base:
                print(
                    FMT_RED
                    + FMT_BOLD
                    + "WARNING: memory section {} in {} is placed before program base (0x{:x})".format(
                        sec, self.name, self.base
                    )
                    + FMT_NONE
                )

        self.sections.append(sec)
        # keep section list ordered
        self.sections.sort(key=lambda section: section.base)

    def __repr__(self):
        return "{}(name={},sections={},base={})".format(
            self.__class__.__name__,
            self.name,
            self.global_labels,
            [s.name for s in self.sections],
            self.base,
        )

    @property
    def entrypoint(self):
        if "_start" in self.context.labels:
            return self.context.labels.get("_start")
        if "main" in self.context.labels:
            return self.context.labels.get("main")
        for sec in self.sections:
            if get_section_base_name(sec.name) == ".text" and sec.flags.executable:
                return sec.base

    def loaded_trigger(self, at_addr: T_AbsoluteAddress):
        """
        This trigger is called when the binary is loaded and its final address in memory is determined

        This will do a small sanity check to prevent programs loading twice, or at addresses they don't
        expect to be loaded.

        Then it will finalize all relative symbols defined in it to point to the correct addresses.

        :param at_addr: the address where the program will be located
        """
        if self.is_loaded:
            if at_addr != self.base:
                raise RuntimeError(
                    "Program loaded twice at different addresses! This will probably break things!"
                )
            return

        if self.base is not None and self.base != at_addr:
            print(
                FMT_MEM
                + "WARNING: Program loaded at different address then expected! (loaded at {}, "
                "but expects to be loaded at {})".format(at_addr, self.base) + FMT_NONE
            )

        # check if we are relocating
        if self.base != at_addr:
            offset = at_addr if self.base is None else at_addr - self.base

            # move all sections by the offset
            for sec in self.sections:
                sec.base += offset

            # move all relative symbols by the offset
            for name in self.relative_labels:
                self.context.labels[name] += offset

        self.base = at_addr
        self.context.base_address = at_addr



import struct
import warnings
from ctypes import c_float, c_double
from typing import Union, Any, ClassVar, Type
from abc import ABC

bytes_t = bytes


class BaseFloat(ABC):
    __slots__ = ("_val",)

    _type: ClassVar[Type[Union[c_float, c_double]]]
    _struct_fmt_str: ClassVar[str]

    _val: Union[c_float, c_double]

    @property
    def value(self) -> float:
        """
        The value represented by this float
        """
        return self._val.value

    @property
    def bytes(self) -> bytes:
        """
        The values bit representation (as a bytes object)
        """
        return struct.pack("<" + self._struct_fmt_str, self.value)

    @classmethod
    def from_bytes(cls, val: Union[bytes_t, bytearray]):
        return cls(val)

    def __init__(
        self, val: Union[float, c_float, "BaseFloat", bytes_t, bytearray, int] = 0
    ):
        if isinstance(val, (float, int)):
            self._val = self._type(val)
        elif isinstance(val, (c_float, c_double)):
            self._val = self._type(val.value)
        elif isinstance(val, (bytes, bytearray)):
            self._val = self._type(struct.unpack("<" + self._struct_fmt_str, val)[0])
        elif isinstance(val, self.__class__):
            self._val = val._val
        else:
            raise ValueError(
                "Unsupported value passed to {}: {} ({})".format(
                    self.__class__.__name__,
                    repr(val),
                    type(val),
                )
            )

    def __add__(self, other: Union["BaseFloat", float]):
        if isinstance(other, BaseFloat):
            other = other.value
        return self.__class__(self.value + other)

    def __sub__(self, other: Union["BaseFloat", float]):
        if isinstance(other, BaseFloat):
            other = other.value
        return self.__class__(self.value - other)

    def __mul__(self, other: Union["BaseFloat", float]):
        if isinstance(other, BaseFloat):
            other = other.value
        return self.__class__(self.value * other)

    def __truediv__(self, other: Any):
        if isinstance(other, BaseFloat):
            other = other.value
        return self.__class__(self.value / other)

    def __floordiv__(self, other: Any):
        if isinstance(other, BaseFloat):
            other = other.value
        return self.__class__(self.value // other)

    def __mod__(self, other: Union["BaseFloat", float]):
        if isinstance(other, BaseFloat):
            other = other.value
        return self.__class__(self.value % other)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, (float, int)):
            return self.value == other
        elif isinstance(other, BaseFloat):
            return self.value == other.value
        return False

    def __neg__(self):
        return self.__class__(-self.value)

    def __abs__(self):
        return self.__class__(abs(self.value))

    def __bytes__(self) -> bytes_t:
        return self.bytes

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.value)

    def __str__(self):
        return str(self.value)

    def __hash__(self):
        return hash(self.value)

    def __gt__(self, other: Any):
        if isinstance(other, BaseFloat):
            other = other.value
        return self.value > other

    def __lt__(self, other: Any):
        if isinstance(other, BaseFloat):
            other = other.value
        return self.value < other

    def __le__(self, other: Any):
        if isinstance(other, BaseFloat):
            other = other.value
        return self.value <= other

    def __ge__(self, other: Any):
        if isinstance(other, BaseFloat):
            other = other.value
        return self.value >= other

    def __bool__(self):
        return bool(self.value)

    def __float__(self):
        return self.value

    def __pow__(self, power, modulo=None):
        if modulo is not None:
            raise ValueError("Float32 pow with modulo unsupported")
        return self.__class__(self.value**power)

    # right handed binary operators

    def __radd__(self, other: Any):
        return self + other

    def __rsub__(self, other: Any):
        return self.__class__(other) - self

    def __rmul__(self, other: Any):
        return self * other

    def __rtruediv__(self, other: Any):
        return self.__class__(other) // self

    def __rfloordiv__(self, other: Any):
        return self.__class__(other) // self

    def __rmod__(self, other: Any):
        return self.__class__(other) % self

    @classmethod
    def bitcast(cls, f: "BaseFloat") -> "BaseFloat":
        """
        bitcast the struct up or down to another type.
        Fills upper bits with zero.

        Use Float64.bitcast(Float32(...)) to bitcast a f32 to f64
        """
        if isinstance(f, cls):
            return f
        return cls.from_bytes(
            (b"\x00\x00\x00\x00\x00\x00\x00\x00" + f.bytes)[
                -struct.calcsize(cls._struct_fmt_str) :
            ]
        )

    @classmethod
    def flen_to_cls(cls, bits: int) -> Type["BaseFloat"]:
        if bits == 32:
            return Float32
        if bits == 64:
            return Float64
        raise ValueError(f"Unsupported flen: {bits}")

    def __format__(self, spec: str):
        if spec[-2:] == "32":
            return Float32.bitcast(self).__format__(spec[:-2])
        if spec[-2:] == "64":
            return Float64.bitcast(self).__format__(spec[:-2])
        return format(self.value, spec)

    def __int__(self):
        warnings.warn(
            "Float32.__int__ is deprecated due its inability to handle infities and NaNs correctly! Use {Int32,UInt32).from_float instead!",
            DeprecationWarning,
        )
        return int(self.value)


class Float32(BaseFloat):
    _type = c_float
    _struct_fmt_str = "f"


class Float64(BaseFloat):
    _type = c_double
    _struct_fmt_str = "d"



from typing import Dict, Tuple

MCAUSE_TRANSLATION: Dict[Tuple[int, int], str] = {
    (1, 0): "User software interrupt",
    (1, 1): "Supervisor software interrupt",
    (1, 3): "Machine software interrupt",
    (1, 4): "User timer interrupt",
    (1, 5): "Supervisor timer interrupt",
    (1, 7): "Machine timer interrupt",
    (1, 8): "User external interrupt",
    (1, 9): "Supervisor external interrupt",
    (1, 11): "Machine external interrupt",
    (0, 0): "Instruction address misaligned",
    (0, 1): "Instruction access fault",
    (0, 2): "Illegal instruction",
    (0, 3): "Breakpoint",
    (0, 4): "Load address misaligned",
    (0, 5): "Load access fault",
    (0, 6): "Store/AMO address misaligned",
    (0, 7): "Store/AMO access fault",
    (0, 8): "environment call from user mode",
    (0, 9): "environment call from supervisor mode",
    (0, 11): "environment call from machine mode",
    (0, 12): "Instruction page fault",
    (0, 13): "Load page fault",
    (0, 15): "Store/AMO page fault",
}
"""
Assigns tuple (interrupt bit, exception code) to their respective readable names
"""

CSR_NAME_TO_ADDR: Dict[str, int] = {
    "fflags": 0x001,
    "frm": 0x002,
    "fcsr": 0x003,
    "mstatus": 0x300,
    "misa": 0x301,
    "mie": 0x304,
    "mtvec": 0x305,
    "mepc": 0x341,
    "mcause": 0x342,
    "mtval": 0x343,
    "mip": 0x344,
    "mtimecmp": 0x780,
    "mtimecmph": 0x781,
    "halt": 0x789,
    "cycle": 0xC00,
    "time": 0xC01,
    "instret": 0xC02,
    "cycleh": 0xC80,
    "timeh": 0xC81,
    "instreth": 0xC82,
    "mvendorid": 0xF11,
    "marchid": 0xF12,
    "mimpid": 0xF13,
    "mhartid": 0xF14,
}
"""
Translation for named registers
"""



from typing import Tuple

from .instruction_set import InstructionSet, Instruction
from ..core import Int32, UInt32
from ..core.csr_constants import CSR_NAME_TO_ADDR
from ..helpers import parse_numeric_argument


class Zicsr(InstructionSet):
    def instruction_csrrw(self, ins: Instruction):
        rd, new_val, csr = self._parse_csr_ins(ins)
        if rd != "zero":
            self.regs.set(rd, self.cpu.csr.get(csr))
        self.cpu.csr.set(csr, new_val)

    def instruction_csrrwi(self, ins: Instruction):
        rd, new_val, csr = self.parse_csr_imm_ins(ins)
        if rd != "zero":
            self.regs.set(rd, self.cpu.csr.get(csr))
        self.cpu.csr.set(csr, new_val)

    def instruction_csrrs(self, ins: Instruction):
        rd, bitmask, csr = self._parse_csr_ins(ins)
        val = self.cpu.csr.get(csr)
        if rd != "zero":
            self.regs.set(rd, val)
        if bitmask != 0:
            self.cpu.csr.set(csr, val | bitmask)

    def instruction_csrrsi(self, ins: Instruction):
        rd, bitmask, csr = self.parse_csr_imm_ins(ins)
        val = self.cpu.csr.get(csr)
        if rd != "zero":
            self.regs.set(rd, val)
        if bitmask != 0:
            self.cpu.csr.set(csr, val | bitmask)

    def instruction_csrrc(self, ins: Instruction):
        rd, bitmask, csr = self._parse_csr_ins(ins)
        val = self.cpu.csr.get(csr)
        if rd != "zero":
            self.regs.set(rd, val)
        if bitmask != 0:
            self.cpu.csr.set(csr, val & ~bitmask)

    def instruction_csrrci(self, ins: Instruction):
        rd, bitmask, csr = self.parse_csr_imm_ins(ins)
        val = self.cpu.csr.get(csr)
        if rd != "zero":
            self.regs.set(rd, val)
        if bitmask != 0:
            self.cpu.csr.set(csr, val & ~bitmask)

    def instruction_rdtime(self, ins: Instruction):
        rd = ins.get_reg(0)
        self.regs.set(rd, self.cpu.rtclock.get_low32())

    def instruction_rdtimeh(self, ins: Instruction):
        rd = ins.get_reg(0)
        self.regs.set(rd, self.cpu.rtclock.get_hi32())

    # FIXME: rdclycle[h] and rdinstret[h] are not provided yet

    def _parse_csr_ins(self, ins: Instruction) -> Tuple[str, UInt32, int]:
        assert len(ins.args) == 3
        rd = ins.get_reg(0)
        rs = self.regs.get(ins.get_reg(1)).unsigned()

        return rd, rs, self._ins_get_csr_addr(ins)

    def parse_csr_imm_ins(self, ins: Instruction) -> Tuple[str, UInt32, int]:
        assert len(ins.args) == 3
        rd = ins.get_reg(0)
        # make sure we only have 5 bit immediate values here
        rs = ins.get_imm(1).abs_value.unsigned()
        if rs > 0b11111:
            print(f"Warning! more than 5bit immediate value in csr ins {ins}!")
            rs = rs & 0b11111

        return rd, rs, self._ins_get_csr_addr(ins)

    @staticmethod
    def _ins_get_csr_addr(ins: Instruction) -> int:
        # TODO: somehow handle elf instructions at some point?
        if isinstance(ins.args[2], str) and ins.args[2].lower() in CSR_NAME_TO_ADDR:
            return CSR_NAME_TO_ADDR[ins.args[2].lower()]

        return ins.get_imm(2).abs_value.unsigned_value



import os.path
from io import IOBase, RawIOBase
from typing import List

from core.traps import *
from .types import ElfMemorySection
from ..helpers import FMT_PARSE, FMT_NONE, FMT_GREEN, FMT_BOLD
from ..core import MemoryFlags, Program, ProgramLoader, T_ParserOpts

FMT_ELF = FMT_GREEN + FMT_BOLD

if typing.TYPE_CHECKING:
    from elftools.elf.elffile import ELFFile
    from elftools.elf.sections import Section, SymbolTableSection

INCLUDE_SEC = (".text", ".stack", ".bss", ".sdata", ".sbss")


class ElfBinaryFileLoader(ProgramLoader):
    """
    Loads compiled elf binaries (checks for the magic sequence 7f45 4c46)

    This loader respects local and global symbols.
    """

    is_binary = True

    program: Program

    def __post_init__(self):
        self.program = Program(self.filename)

    @classmethod
    def can_parse(cls, source_name: str) -> float:
        if source_name == "-" or not os.path.isfile(source_name):
            return 0

        with open(source_name, "rb") as f:
            if f.read(4) == b"\x7f\x45\x4c\x46":
                return 1
        return 0

    @classmethod
    def get_options(cls, argv: List[str]) -> [List[str], T_ParserOpts]:
        return argv, {}

    def parse(self) -> Program:
        try:
            from elftools.elf.elffile import ELFFile
            from elftools.elf.sections import Section, SymbolTableSection

            self._read_elf(ELFFile(self.source))
        except ImportError as e:
            print(
                FMT_PARSE
                + "[ElfLoader] Cannot load elf files without PyElfTools package! You can install them "
                "using pip install pyelftools!" + FMT_NONE
            )
            raise e

        return self.program

    def _read_elf(self, elf: "ELFFile"):
        if not elf.header.e_machine == "EM_RISCV":
            raise InvalidElfException("Not a RISC-V elf file!")
        if not elf.header.e_ident.EI_CLASS == "ELFCLASS32":
            raise InvalidElfException("Only 32bit executables are supported!")

        from elftools.elf.sections import SymbolTableSection

        for sec in elf.iter_sections():
            if isinstance(sec, SymbolTableSection):
                self._parse_symtab(sec)
                continue

            if sec.name not in INCLUDE_SEC:
                continue

            self._add_sec(self._lms_from_elf_sec(sec, self.filename))

    def _lms_from_elf_sec(self, sec: "Section", owner: str):
        is_code = sec.name in (".text",)
        data = bytearray(sec.data())
        if len(data) < sec.data_size:
            data += bytearray(len(data) - sec.data_size)
        flags = MemoryFlags(is_code, is_code)
        print(
            FMT_ELF
            + "[ElfLoader] Section {} at: {:X}".format(sec.name, sec.header.sh_addr)
            + FMT_NONE
        )
        return ElfMemorySection(
            data, sec.name, self.program.context, owner, sec.header.sh_addr, flags
        )

    def _parse_symtab(self, symtab: "SymbolTableSection"):
        for sym in symtab.iter_symbols():
            if not sym.name:
                continue
            self.program.context.labels[sym.name] = sym.entry.st_value
            # check if it has st_visibility bit set
            if sym.entry.st_info.bind == "STB_GLOBAL":
                self.program.global_labels.add(sym.name)
                print(
                    FMT_PARSE
                    + "LOADED GLOBAL SYMBOL {}: {}".format(sym.name, sym.entry.st_value)
                    + FMT_NONE
                )

    def _add_sec(self, new_sec: "ElfMemorySection"):
        for sec in self.program.sections:
            if sec.base < sec.end <= new_sec.base or sec.end > sec.base >= new_sec.end:
                continue
            else:
                print(
                    FMT_ELF
                    + "[ElfLoader] Invalid elf layout: Two sections overlap: \n\t{}\n\t{}".format(
                        sec, new_sec
                    )
                    + FMT_NONE
                )
                raise RuntimeError("Cannot load elf with overlapping sections!")

        self.program.add_section(new_sec)



"""
RiscEmu (c) 2023 Anton Lydike

SPDX-License-Identifier: MIT

This file contains copious amounts of docstrings that were all taken
from https://msyksphinz-self.github.io/riscv-isadoc/html/rvfd.html
(all the docstrings on the instruction methods documenting the opcodes
and their function)
"""
from .instruction_set import Instruction
from .float_base import FloatArithBase
from riscemu.core import Float32, Int32, UInt32


class RV32F(FloatArithBase[Float32]):
    flen = 32
    _float_cls = Float32

    def instruction_fcvt_w_s(self, ins: Instruction):
        """
        +-----+-----+-----+-----+-----+-----+-----+---+
        |31-27|26-25|24-20|19-15|14-12|11-7 |6-2  |1-0|
        +-----+-----+-----+-----+-----+-----+-----+---+
        |11000|00   |00000|rs1  |rm   |rd   |10100|11 |
        +-----+-----+-----+-----+-----+-----+-----+---+

        :Format:
        | fcvt.w.s   rd,rs1

        :Description:
        | Convert a floating-point number in floating-point register rs1 to a signed 32-bit in integer register rd.

        :Implementation:
        | x[rd] = sext(s32_{f32}(f[rs1]))
        """
        rd, rs = self.parse_rd_rs(ins)
        self.regs.set(rd, Int32.from_float(self.regs.get_f(rs).value))

    def instruction_fcvt_wu_s(self, ins: Instruction):
        """
        +-----+-----+-----+-----+-----+-----+-----+---+
        |31-27|26-25|24-20|19-15|14-12|11-7 |6-2  |1-0|
        +-----+-----+-----+-----+-----+-----+-----+---+
        |11000|00   |00001|rs1  |rm   |rd   |10100|11 |
        +-----+-----+-----+-----+-----+-----+-----+---+

        :Format:
        | fcvt.wu.s  rd,rs1

        :Description:
        | Convert a floating-point number in floating-point register rs1 to a signed 32-bit in unsigned integer register rd.

        :Implementation:
        | x[rd] = sext(u32_{f32}(f[rs1]))
        """
        rd, rs = self.parse_rd_rs(ins)
        self.regs.set(rd, UInt32.from_float((self.regs.get_f(rs).value)))

    def instruction_fmv_x_w(self, ins: Instruction):
        """
        +-----+-----+-----+-----+-----+-----+-----+---+
        |31-27|26-25|24-20|19-15|14-12|11-7 |6-2  |1-0|
        +-----+-----+-----+-----+-----+-----+-----+---+
        |11100|00   |00000|rs1  |000  |rd   |10100|11 |
        +-----+-----+-----+-----+-----+-----+-----+---+

        :Format:
        | fmv.x.w    rd,rs1

        :Description:
        | Move the single-precision value in floating-point register rs1 represented in IEEE 754-2008 encoding to the lower 32 bits of integer register rd.

        :Implementation:
        | x[rd] = sext(f[rs1][31:0])
        """
        rd, rs = self.parse_rd_rs(ins)
        self.regs.set(rd, UInt32(self.regs.get_f(rs).bytes[-4:]))

    def instruction_fcvt_s_w(self, ins: Instruction):
        """
        +-----+-----+-----+-----+-----+-----+-----+---+
        |31-27|26-25|24-20|19-15|14-12|11-7 |6-2  |1-0|
        +-----+-----+-----+-----+-----+-----+-----+---+
        |11010|00   |00000|rs1  |rm   |rd   |10100|11 |
        +-----+-----+-----+-----+-----+-----+-----+---+

        :Format:
        | fcvt.s.w   rd,rs1

        :Description:
        | Converts a 32-bit signed integer, in integer register rs1 into a floating-point number in floating-point register rd.

        :Implementation:
        | f[rd] = f32_{s32}(x[rs1])
        """
        rd, rs = self.parse_rd_rs(ins)
        self.regs.set_f(rd, Float32(self.regs.get(rs).value))

    def instruction_fcvt_s_wu(self, ins: Instruction):
        """
        +-----+-----+-----+-----+-----+-----+-----+---+
        |31-27|26-25|24-20|19-15|14-12|11-7 |6-2  |1-0|
        +-----+-----+-----+-----+-----+-----+-----+---+
        |11010|00   |00001|rs1  |rm   |rd   |10100|11 |
        +-----+-----+-----+-----+-----+-----+-----+---+

        :Format:
        | fcvt.s.wu  rd,rs1

        :Description:
        | Converts a 32-bit unsigned integer, in integer register rs1 into a floating-point number in floating-point register rd.

        :Implementation:
        | f[rd] = f32_{u32}(x[rs1])
        """
        rd, rs = self.parse_rd_rs(ins)
        self.regs.set_f(rd, Float32(self.regs.get(rs).unsigned_value))

    def instruction_fmv_w_x(self, ins: Instruction):
        """
        +-----+-----+-----+-----+-----+-----+-----+---+
        |31-27|26-25|24-20|19-15|14-12|11-7 |6-2  |1-0|
        +-----+-----+-----+-----+-----+-----+-----+---+
        |11110|00   |00000|rs1  |000  |rd   |10100|11 |
        +-----+-----+-----+-----+-----+-----+-----+---+



        :Format:
        | fmv.w.x    rd,rs1

        :Description:
        | Move the single-precision value encoded in IEEE 754-2008 standard encoding from the lower 32 bits of integer register rs1 to the floating-point register rd.

        :Implementation:
        | f[rd] = x[rs1][31:0]
        """
        rd, rs = self.parse_rd_rs(ins)
        self.regs.set_f(rd, Float32.from_bytes(self.regs.get(rs).to_bytes()))



from .types import ElfMemorySection
from core.mmu import *

import typing

if typing.TYPE_CHECKING:
    pass


class PrivMMU(MMU):
    def get_sec_containing(self, addr: T_AbsoluteAddress) -> MemorySection:
        # try to get an existing section
        existing_sec = super().get_sec_containing(addr)

        if existing_sec is not None:
            return existing_sec

        # get section preceding empty space at addr
        sec_before = next(
            (sec for sec in reversed(self.sections) if sec.end < addr), None
        )
        # get sec succeeding empty space at addr
        sec_after = next((sec for sec in self.sections if sec.base > addr), None)

        # calc start and end of "free" space
        prev_sec_end = 0 if sec_before is None else sec_before.end
        next_sec_start = 0x7FFFFFFF if sec_after is None else sec_after.base

        # start at the end of the prev section, or current address - 0xFFFF (aligned to 16 byte boundary)
        start = max(prev_sec_end, align_addr(addr - 0xFFFF, 16))
        # end at the start of the next section, or address + 0xFFFF (aligned to 16 byte boundary)
        end = min(next_sec_start, align_addr(addr + 0xFFFF, 16))

        sec = ElfMemorySection(
            bytearray(end - start),
            ".empty",
            self.global_instruction_context(),
            "",
            start,
            MemoryFlags(False, True),
        )
        self.sections.append(sec)
        self._update_state()

        return sec

    def global_instruction_context(self) -> InstructionContext:
        context = InstructionContext()
        context.global_symbol_dict = self.global_symbols
        return context



"""
RiscEmu (c) 2021 Anton Lydike

SPDX-License-Identifier: MIT
"""

from math import log10, ceil
from typing import Iterable, Iterator, TypeVar, Generic, List, Optional

from .core import Int32, UInt32
from .core.exceptions import *


def align_addr(addr: int, to_bytes: int = 8) -> int:
    """
    align an address to `to_bytes` (meaning addr & to_bytes = 0)

    This will increase the address
    """
    return addr + (-addr % to_bytes)


def parse_numeric_argument(arg: str) -> int:
    """
    parse hex or int strings
    """
    try:
        if arg.lower().startswith("0x"):
            return int(arg, 16)
        return int(arg)
    except ValueError as ex:
        raise ParseException(
            'Invalid immediate argument "{}", maybe missing symbol?'.format(arg),
            (arg, ex),
        )


def create_chunks(my_list, chunk_size):
    """Split a list like [a,b,c,d,e,f,g,h,i,j,k,l,m] into e.g. [[a,b,c,d],[e,f,g,h],[i,j,k,l],[m]]"""
    return [my_list[i : i + chunk_size] for i in range(0, len(my_list), chunk_size)]


def apply_highlight(item, ind, hi_ind):
    """
    applies some highlight such as underline to item if ind == hi_ind
    """
    if ind == hi_ind:
        return FMT_UNDERLINE + FMT_ORANGE + item + FMT_NONE
    return item


def highlight_in_list(items, hi_ind, joiner=" "):
    return joiner.join(
        [apply_highlight(item, i, hi_ind) for i, item in enumerate(items)]
    )


def format_bytes(byte_arr: bytearray, fmt: str, group: int = 1, highlight: int = -1):
    """Format byte array as per fmt. Group into groups of size `group`, and highlight index `highlight`."""
    chunks = create_chunks(byte_arr, group)
    if fmt == "hex":
        return highlight_in_list(["0x{}".format(ch.hex()) for ch in chunks], highlight)
    if fmt == "int":
        spc = str(ceil(log10(2 ** (group * 8 - 1))) + 1)
        return highlight_in_list(
            [("{:0" + spc + "d}").format(Int32(ch)) for ch in chunks], highlight
        )
    if fmt == "uint":
        spc = str(ceil(log10(2 ** (group * 8))))
        return highlight_in_list(
            [("{:0" + spc + "d}").format(UInt32(ch)) for ch in chunks], highlight
        )
    if fmt == "char":
        return highlight_in_list((repr(chr(b))[1:-1] for b in byte_arr), highlight, "")


T = TypeVar("T")


class Peekable(Generic[T], Iterator[T]):
    def __init__(self, iterable: Iterable[T]):
        self.iterable = iter(iterable)
        self.cache: List[T] = list()

    def __iter__(self) -> Iterator[T]:
        return self

    def __next__(self) -> T:
        if self.cache:
            return self.cache.pop()
        return next(self.iterable)

    def peek(self) -> Optional[T]:
        try:
            if self.cache:
                return self.cache[0]
            pop = next(self.iterable)
            self.cache.append(pop)
            return pop
        except StopIteration:
            return None

    def push_back(self, item: T):
        self.cache = [item] + self.cache

    def is_empty(self) -> bool:
        return self.peek() is None


def get_section_base_name(section_name: str) -> str:
    if "." not in section_name:
        print(
            FMT_PARSE
            + f"Invalid section {section_name}, not starting with a dot!"
            + FMT_NONE
        )
    return "." + section_name.split(".")[1]



"""
RiscEmu (c) 2021 Anton Lydike

SPDX-License-Identifier: MIT
"""
import sys
import time

from core.usermode_cpu import *
from .CSR import CSR
from .ElfLoader import ElfBinaryFileLoader
from core.traps import *
from .ImageLoader import MemoryImageLoader
from .PrivMMU import PrivMMU
from .PrivRV32I import PrivRV32I
from core.privmodes import PrivModes
from ..IO.TextIO import TextIO
from ..instructions import RV32A, RV32M
from ..core import Program, UInt32

if typing.TYPE_CHECKING:
    pass


class PrivCPU(CPU):
    """
    This is a CPU that has different modes, instruction sets and registers.

    It should support M and U Mode, but no U-Mode Traps.

    This is meant to emulate whole operating systems.
    """

    csr: CSR
    """
    Reference to the control and status registers
    """

    TIME_RESOLUTION_NS: int = 10000000
    """
    controls the resolution of the time csr register (in nanoseconds)
    """

    pending_traps: List[CpuTrap]
    """
    A list of traps which are pending to be handled
    """

    def __init__(self, conf):
        super().__init__(PrivMMU(), [PrivRV32I, RV32M, RV32A], conf)
        # start in machine mode
        self.mode: PrivModes = PrivModes.MACHINE

        self.pending_traps: List[CpuTrap] = list()

        self.exit_code = 0

        self._time_start = 0
        self._time_timecmp = UInt32(0)
        self._time_interrupt_enabled = False

        # performance counters
        self._perf_counters = list()

        # add TextIO
        io = TextIO(0xFF0000, 64)
        self.mmu.load_section(io, True)

        # init csr
        self._init_csr()

        self.TIME_RESOLUTION_NS = int(self.TIME_RESOLUTION_NS * conf.slowdown)

    def run(self, verbose=False):
        if self.pc <= 0:
            return False

        launch_debug = False

        try:
            while not self.halted:
                self.step(verbose)
        except RiscemuBaseException as ex:
            if isinstance(ex, LaunchDebuggerException):
                launch_debug = True
                self.pc += self.INS_XLEN

        if self.halted:
            print()
            print(
                FMT_CPU
                + "[CPU] System halted with code {}".format(self.exit_code)
                + FMT_NONE
            )
            sys.exit(self.exit_code)

        elif launch_debug:
            launch_debug_session(self)
            if not self.debugger_active:
                self.run(verbose)
        else:
            print()
            print(
                FMT_CPU
                + "[CPU] System stopped without halting - perhaps you stopped the debugger?"
                + FMT_NONE
            )

    def launch(self, program: Optional[Program] = None, verbose: bool = False):
        print(
            FMT_CPU
            + "[CPU] Started running from 0x{:08X} ({})".format(self.pc, "kernel")
            + FMT_NONE
        )
        self._time_start = time.perf_counter_ns() // self.TIME_RESOLUTION_NS

        self.run(self.conf.verbosity > 1 or verbose)

    def load_program(self, program: Program):
        if program.name == "kernel":
            self.pc = program.entrypoint
        super().load_program(program)

    def _init_csr(self):
        # set up CSR
        self.csr = CSR()
        self.csr.set("mhartid", UInt32(0))  # core id
        # TODO: set correct value
        self.csr.set("mimpid", UInt32(0))  # implementation id
        # set mxl to 1 (32 bit) and set bits for i and m isa
        self.csr.set("misa", UInt32((1 << 30) + (1 << 8) + (1 << 12)))  # available ISA

        # CSR write callbacks:

        @self.csr.callback("halt")
        def halt(old: UInt32, new: UInt32):
            if new != 0:
                self.halted = True
                self.exit_code = new.value

        @self.csr.callback("mtimecmp")
        def mtimecmp(old: UInt32, new: UInt32):
            self._time_timecmp = (self.csr.get("mtimecmph") << 32) + new
            self._time_interrupt_enabled = True

        @self.csr.callback("mtimecmph")
        def mtimecmph(old: UInt32, new: UInt32):
            self._time_timecmp = (new << 32) + self.csr.get("mtimecmp")
            self._time_interrupt_enabled = True

        # virtual CSR registers:

        @self.csr.virtual_register("time")
        def get_time():
            return UInt32(
                time.perf_counter_ns() // self.TIME_RESOLUTION_NS - self._time_start
            )

        @self.csr.virtual_register("timeh")
        def get_timeh():
            return UInt32(
                (time.perf_counter_ns() // self.TIME_RESOLUTION_NS - self._time_start)
                >> 32
            )

        # add minstret and mcycle counters

    def _handle_trap(self, trap: CpuTrap):
        # implement trap handling!
        self.pending_traps.append(trap)

    def step(self, verbose=True):
        try:
            self.cycle += 1
            if self.cycle % 20 == 0:
                self._timer_step()
            self._check_interrupt()
            ins = self.mmu.read_ins(self.pc)
            if verbose and (self.mode == PrivModes.USER or self.conf.verbosity > 4):
                print(
                    FMT_CPU + "   Running 0x{:08X}:{} {}".format(self.pc, FMT_NONE, ins)
                )
            self.run_instruction(ins)
            self.pc += self.INS_XLEN
        except CpuTrap as trap:
            self._handle_trap(trap)
            if trap.interrupt == 0 and not isinstance(trap, EcallTrap):
                print(
                    FMT_CPU
                    + "[CPU] Trap {} encountered at {} (0x{:x})".format(
                        trap, self.mmu.translate_address(self.pc), self.pc
                    )
                    + FMT_NONE
                )
                breakpoint()
                if self.conf.debug_on_exception:
                    raise LaunchDebuggerException()
            self.pc += self.INS_XLEN

    def _timer_step(self):
        if not self._time_interrupt_enabled:
            return
        if (
            self._time_timecmp
            <= (time.perf_counter_ns() // self.TIME_RESOLUTION_NS) - self._time_start
        ):
            self.pending_traps.append(TimerInterrupt())
            self._time_interrupt_enabled = False

    def _check_interrupt(self):
        if not (len(self.pending_traps) > 0 and self.csr.get_mstatus("mie")):
            return
        # select best interrupt
        # FIXME: actually select based on the official ranking
        trap = self.pending_traps.pop()  # use the most recent trap
        if self.conf.verbosity > 0:
            print(FMT_CPU + "[CPU] taking trap {}!".format(trap) + FMT_NONE)
            self.regs.dump_reg_a()

        if trap.priv != PrivModes.MACHINE:
            print(
                FMT_CPU
                + "[CPU] Trap not targeting machine mode encountered! - undefined behaviour!"
                + FMT_NONE
            )
            raise Exception("Undefined behaviour!")

        if self.mode != PrivModes.USER:
            print(FMT_CPU + "[CPU] Trap triggered outside of user mode?!" + FMT_NONE)

        self.csr.set_mstatus("mpie", self.csr.get_mstatus("mie"))
        self.csr.set_mstatus("mpp", self.mode.value)
        self.csr.set_mstatus("mie", UInt32(0))
        self.csr.set("mcause", trap.mcause)
        self.csr.set("mepc", self.pc - self.INS_XLEN)
        self.csr.set("mtval", trap.mtval)
        self.mode = trap.priv
        mtvec = self.csr.get("mtvec")
        if mtvec & 0b11 == 0:
            self.pc = mtvec.value
        if mtvec & 0b11 == 1:
            self.pc = (
                (mtvec & 0b11111111111111111111111111111100) + (trap.code * 4)
            ).value
        self.record_perf_profile()
        if len(self._perf_counters) > 100:
            self.show_perf()

    def show_perf(self):
        timed = 0
        cycled = 0
        cps_list = list()

        print(FMT_CPU + "[CPU] Performance overview:")
        for time_ns, cycle in self._perf_counters:
            if cycled == 0:
                cycled = cycle
                timed = time_ns
                continue
            cps = (cycle - cycled) / (time_ns - timed) * 1000000000

            cycled = cycle
            timed = time_ns
            cps_list.append(cps)
        print(
            "    on average {:.0f} instructions/s".format(sum(cps_list) / len(cps_list))
            + FMT_NONE
        )
        self._perf_counters = list()

    def record_perf_profile(self):
        self._perf_counters.append((time.perf_counter_ns(), self.cycle))



"""
RiscEmu (c) 2021 Anton Lydike

SPDX-License-Identifier: MIT
"""

from typing import Dict, List, Optional, Union

from ..colors import *
from ..helpers import align_addr
from . import (
    Instruction,
    MemorySection,
    MemoryFlags,
    T_AbsoluteAddress,
    Program,
    InstructionContext,
    Int32,
    Float32,
    Float64,
    InvalidAllocationException,
    MemoryAccessException,
)


class MMU:
    """
    The MemoryManagementUnit. This provides a unified interface for reading/writing data from/to memory.

    It also provides various translations for addresses.
    """

    max_size = 0xFFFFFFFF
    """
    The maximum size of the memory in bytes
    """

    max_alloc_size = 8 * 1024 * 1024 * 64
    """
    No single allocation can be bigger than 64 MB
    """

    sections: List[MemorySection]
    """
    A list of all loaded memory sections
    """

    programs: List[Program]
    """
    A list of all loaded programs
    """

    global_symbols: Dict[str, int]
    """
    The global symbol table
    """

    _ins_sec: Optional[MemorySection]
    """
    Caching the last section where we read instructions from
    """

    _mem_sec: Optional[MemorySection]
    """
    Caching the last section where we read data from
    """

    def __init__(self):
        """
        Create a new MMU
        """
        self.programs = list()
        self.sections = list()
        self.global_symbols = dict()
        self._ins_sec = None
        self._mem_sec = None

    def get_sec_containing(self, addr: T_AbsoluteAddress) -> Optional[MemorySection]:
        """
        Returns the section that contains the address addr

        :param addr: the Address to look for
        :return: The LoadedMemorySection or None
        """
        for sec in self.sections:
            if sec.base <= addr < sec.base + sec.size:
                self._mem_sec = sec
                return sec
        return None

    def get_program_at_addr(self, addr: T_AbsoluteAddress) -> Optional[Program]:
        for program in self.programs:
            if program.base <= addr < program.base + program.size:
                return program
        return None

    def read_ins(self, addr: T_AbsoluteAddress) -> Instruction:
        """
        Read a single instruction located at addr

        :param addr: The location
        :return: The Instruction
        """
        sec = self._ins_sec
        if addr < sec.base or sec.base + sec.size <= addr:
            sec = self.get_sec_containing(addr)
            if sec is not None:
                self._ins_sec = sec
            else:
                print(
                    FMT_MEM
                    + "[MMU] Trying to read instruction form invalid region! (read at {}) ".format(
                        addr
                    )
                    + "Have you forgotten an exit syscall or ret statement?"
                    + FMT_NONE
                )
                raise RuntimeError("No next instruction available!")
        return sec.read_ins(addr - sec.base)

    def read(self, addr: Union[int, Int32], size: int) -> bytearray:
        """
        Read size bytes of memory at addr

        :param addr: The address at which to start reading
        :param size: The number of bytes to read
        :return: The bytearray at addr
        """
        sec = self._mem_sec
        if addr < sec.base or sec.base + sec.size <= addr:
            sec = self.get_sec_containing(addr)
            if sec is not None:
                self._mem_sec = sec
            else:
                print(
                    FMT_MEM
                    + "[MMU] Trying to read data form invalid region at 0x{:x}! ".format(
                        addr
                    )
                    + FMT_NONE
                )
                raise MemoryAccessException(
                    "region is non-initialized!", addr, size, "read"
                )
        return sec.read(addr - sec.base, size)

    def write(self, addr: int, size: int, data: bytearray):
        """
        Write bytes into memory

        :param addr: The address at which to write
        :param size: The number of bytes to write
        :param data: The bytearray to write (only first size bytes are written)
        """
        sec = self._mem_sec
        if addr < sec.base or sec.base + sec.size <= addr:
            sec = self.get_sec_containing(addr)
            if sec is not None:
                self._mem_sec = sec
            else:
                print(
                    FMT_MEM
                    + "[MMU] Invalid write into non-initialized region at 0x{:08X}".format(
                        addr
                    )
                    + FMT_NONE
                )
                raise MemoryAccessException(
                    "region is non-initialized!", addr, size, "write"
                )

        return sec.write(addr - sec.base, size, data)

    def dump(self, addr, *args, **kwargs):
        """
        Dumpy the memory contents

        :param addr: The address at which to dump
        :param args: args for the dump function of the loaded memory section
        :param kwargs: kwargs for the dump function of the loaded memory section
        """
        sec = self.get_sec_containing(addr)
        if sec is None:
            print(
                FMT_MEM
                + "[MMU] No section containing addr 0x{:08X}".format(addr)
                + FMT_NONE
            )
            return
        sec.dump(addr - sec.base, *args, **kwargs)

    def label(self, symb: str):
        """
        Look up the symbol symb in all local symbol tables (and the global one)

        :param symb: The symbol name to look up
        """
        print(FMT_MEM + "[MMU] Lookup for symbol {}:".format(symb) + FMT_NONE)
        if symb in self.global_symbols:
            print(
                "   Found global symbol {}: 0x{:X}".format(
                    symb, self.global_symbols[symb]
                )
            )
        for bin in self.programs:
            if symb in bin.context.labels:
                print(
                    "   Found local labels {}: 0x{:X} in {}".format(
                        symb, bin.context.labels[symb], bin.name
                    )
                )

    def read_int(self, addr: int) -> Int32:
        return Int32(self.read(addr, 4))

    def read_float(self, addr: int) -> Float32:
        return Float32(self.read(addr, 4))

    def read_double(self, addr: int) -> Float64:
        return Float64(self.read(addr, 8))

    def translate_address(self, address: T_AbsoluteAddress) -> str:
        sec = self.get_sec_containing(address)
        if not sec:
            return "unknown at 0x{:0x}".format(address)

        bin = self.get_program_at_addr(address)
        secs = set(sec.name for sec in bin.sections) if bin else []
        elf_markers = {
            "__global_pointer$",
            "_fdata",
            "_etext",
            "_gp",
            "_bss_start",
            "_bss_end",
            "_ftext",
            "_edata",
            "_end",
            "_fbss",
        }

        def key(x):
            name, val = x
            return address - val

        best_fit = sorted(
            filter(lambda x: x[1] <= address, sec.context.labels.items()), key=key
        )

        best = ("", float("inf"))
        for name, val in best_fit:
            if address - val < best[1]:
                best = (name, address - val)
            if address - val == best[1]:
                if best[0] in elf_markers:
                    best = (name, val)
                elif best[0] in secs and name not in elf_markers:
                    best = (name, val)

        name, val = best

        if not name:
            return "{}:{} + 0x{:x} (0x{:x})".format(
                sec.owner, sec.name, address - sec.base, address
            )

        if val == 0:
            return "{}:{} {}".format(sec.owner, sec.name, name)

        return "{}:{} at {} (0x{:0x}) + 0x{:0x}".format(
            sec.owner, sec.name, name, address + val, val
        )

    def has_continuous_free_region(self, start: int, end: int) -> bool:
        # if we have no sections we are all good
        if len(self.sections) == 0:
            return True
        # if the last section is located before the start we are also good
        if start >= self.sections[-1].base + self.sections[-1].size:
            return True

        for sec in self.sections:
            # skip all sections that end before the required start point
            if sec.base + sec.size <= start:
                continue
            # we now have the first section that doesn't end **before** the start point
            # if this section starts after the specified end, we are good
            if sec.base >= end:
                return True
            # otherwise we can't continue
            return False
        # if all sections end before the requested start we are good
        # technically we shouldn't ever reach this point, but better safe than sorry
        return True

    def load_program(self, program: Program, align_to: int = 4):
        if program.base is not None:
            if not self.has_continuous_free_region(
                program.base, program.base + program.size
            ):
                print(
                    FMT_MEM
                    + "Cannot load program {} into desired space (0x{:0x}-0x{:0x}), area occupied.".format(
                        program.name, program.base, program.base + program.size
                    )
                    + FMT_NONE
                )
                raise InvalidAllocationException(
                    "Area occupied".format(
                        program.name, program.base, program.base + program.size
                    ),
                    program.name,
                    program.size,
                    MemoryFlags(False, True),
                )

            at_addr = program.base
        else:
            at_addr = align_addr(self.get_guaranteed_free_address(), align_to)

        # trigger the load event to set all addresses in the binary
        program.loaded_trigger(at_addr)

        # add program and sections to internal state
        self.programs.append(program)
        self.sections += program.sections
        self._update_state()

        # load all global symbols from program
        self.global_symbols.update(
            {key: program.context.labels[key] for key in program.global_labels}
        )
        # inject reference to global symbol table into program context
        # FIXME: this is pretty unclean and should probably be solved in a better way in the future
        program.context.global_symbol_dict = self.global_symbols

    def load_section(self, sec: MemorySection, fixed_position: bool = False) -> bool:
        if fixed_position:
            if self.has_continuous_free_region(sec.base, sec.base + sec.size):
                self.sections.append(sec)
                self._update_state()
            else:
                print(
                    FMT_MEM
                    + "[MMU] Cannot place section {} at {}, space is occupied!".format(
                        sec, sec.base
                    )
                )
                return False
        else:
            at_addr = align_addr(self.get_guaranteed_free_address(), 8)
            sec.base = at_addr
            self.sections.append(sec)
            self._update_state()
        return True

    def _update_state(self):
        """
        Called whenever a section or program is added to keep the list of programs and sections consistent
        :return:
        """
        self.programs.sort(key=lambda bin: bin.base)
        self.sections.sort(key=lambda sec: sec.base)
        self._mem_sec = self.sections[-1]
        self._ins_sec = self.sections[-1]

    def get_guaranteed_free_address(self) -> T_AbsoluteAddress:
        if len(self.sections) == 0:
            return 0x100
        else:
            return self.sections[-1].base + self.sections[-1].size

    def __repr__(self):
        return "{}(\n\t{}\n)".format(
            self.__class__.__name__, "\n\t".join(repr(x) for x in self.programs)
        )

    def context_for(self, addr: T_AbsoluteAddress) -> InstructionContext:
        sec = self.get_sec_containing(addr)

        if sec is not None:
            return sec.context

        return InstructionContext()

    def find_entrypoint(self) -> Optional[int]:
        # try to find the global entrypoint
        if "_start" in self.global_symbols:
            return self.global_symbols["_start"]
        # otherwise find a main (that's not necessarily global)
        for p in self.programs:
            if "main" in p.context.labels:
                return p.context.resolve_label("main")
        return None



from typing import Optional
from . import (
    MemorySection,
    InstructionContext,
    MemoryFlags,
    T_RelativeAddress,
    Instruction,
)
from ..core.exceptions import MemoryAccessException


class BinaryDataMemorySection(MemorySection):
    def __init__(
        self,
        data: bytearray,
        name: str,
        context: InstructionContext,
        owner: str,
        base: int = 0,
        flags: Optional[MemoryFlags] = None,
    ):
        super().__init__(
            name,
            flags if flags is not None else MemoryFlags(False, False),
            len(data),
            base,
            owner,
            context,
        )
        self.data = data

    def read(self, offset: T_RelativeAddress, size: int) -> bytearray:
        if offset + size > self.size:
            raise MemoryAccessException(
                "Out of bounds access in {}".format(self), offset, size, "read"
            )
        return self.data[offset : offset + size]

    def write(self, offset: T_RelativeAddress, size: int, data: bytearray):
        if offset + size > self.size:
            raise MemoryAccessException(
                "Out of bounds access in {}".format(self), offset, size, "write"
            )
        if len(data[0:size]) != size:
            raise MemoryAccessException(
                "Invalid write parameter sizing", offset, size, "write"
            )
        self.data[offset : offset + size] = data[0:size]

    def read_ins(self, offset: T_RelativeAddress) -> Instruction:
        raise MemoryAccessException(
            "Tried reading instruction on non-executable section {}".format(self),
            offset,
            4,
            "instruction fetch",
        )



from typing import Dict, Union, Callable, Optional
from collections import defaultdict
from core.privmodes import PrivModes
from core.traps import InstructionAccessFault
from ..colors import FMT_CSR, FMT_NONE

from core.csr_constants import CSR_NAME_TO_ADDR, MSTATUS_LEN_2, MSTATUS_OFFSETS
from ..core import UInt32


class CSR:
    """
    This holds all Control and Status Registers (CSR)
    """

    regs: Dict[int, UInt32]
    """
    All Control and Status Registers are stored here
    """

    virtual_regs: Dict[int, Callable[[], UInt32]]
    """
    list of virtual CSR registers, with values computed on read
    """

    listeners: Dict[int, Callable[[UInt32, UInt32], None]]

    mstatus_cache: Dict[str, UInt32]
    mstatus_cache_dirty = True

    def __init__(self):
        self.regs = defaultdict(lambda: UInt32(0))
        self.listeners = defaultdict(lambda: (lambda x, y: None))
        self.virtual_regs = dict()
        self.mstatus_cache = dict()
        # TODO: implement write masks (bitmasks which control writeable bits in registers

    def set(self, addr: Union[str, int], val: Union[int, UInt32]):
        addr = self._name_to_addr(addr)
        if addr is None:
            return
        val = UInt32(val)
        self.listeners[addr](self.regs[addr], val)
        if addr == 0x300:
            self.mstatus_cache_dirty = True
        self.regs[addr] = val

    def get(self, addr: Union[str, int]) -> UInt32:
        addr = self._name_to_addr(addr)
        if addr is None:
            raise RuntimeError(f"Invalid CSR name: {addr}!")
        if addr in self.virtual_regs:
            return self.virtual_regs[addr]()
        return self.regs[addr]

    def set_listener(
        self, addr: Union[str, int], listener: Callable[[UInt32, UInt32], None]
    ):
        addr = self._name_to_addr(addr)
        if addr is None:
            print("unknown csr address name: {}".format(addr))
            return
        self.listeners[addr] = listener

    # mstatus properties
    def set_mstatus(self, name: str, val: UInt32):
        """
        Set mstatus bits using this helper. mstatus is a 32 bit register, holding various machine status flags
        Setting them by hand is super painful, so this helper allows you to set specific bits.

        Please make sure your supplied value has the correct width!

        :param name:
        :param val:
        :return:
        """
        size = 2 if name in MSTATUS_LEN_2 else 1
        off = MSTATUS_OFFSETS[name]
        mask = (2**size - 1) << off
        old_val = self.get("mstatus")
        erased = old_val & (~mask)
        new_val = erased | (val << off)
        self.set("mstatus", new_val)

    def get_mstatus(self, name) -> UInt32:
        if not self.mstatus_cache_dirty and name in self.mstatus_cache:
            return self.mstatus_cache[name]

        size = 2 if name in MSTATUS_LEN_2 else 1
        off = MSTATUS_OFFSETS[name]
        mask = (2**size - 1) << off
        val = (self.get("mstatus") & mask) >> off
        if self.mstatus_cache_dirty:
            self.mstatus_cache = dict(name=val)
        else:
            self.mstatus_cache[name] = val
        return val

    def callback(self, addr: Union[str, int]):
        def inner(func: Callable[[UInt32, UInt32], None]):
            self.set_listener(addr, func)
            return func

        return inner

    def assert_can_read(self, mode: PrivModes, addr: int):
        if (addr >> 8) & 3 > mode.value:
            raise InstructionAccessFault(addr)

    def assert_can_write(self, mode: PrivModes, addr: int):
        if (addr >> 8) & 3 > mode.value or addr >> 10 == 11:
            raise InstructionAccessFault(addr)

    def _name_to_addr(self, addr: Union[str, int]) -> Optional[int]:
        if isinstance(addr, str):
            if addr not in CSR_NAME_TO_ADDR:
                print("Unknown CSR register {}".format(addr))
                return None
            return CSR_NAME_TO_ADDR[addr]
        return addr

    def virtual_register(self, addr: Union[str, int]):
        addr = self._name_to_addr(addr)
        if addr is None:
            print("unknown csr address name: {}".format(addr))

        def inner(func: Callable[[], UInt32]):
            self.virtual_regs[addr] = func
            return func

        return inner

    def dump_mstatus(self):
        print(FMT_CSR + "[CSR] dumping mstatus:")
        i = 0
        for name in MSTATUS_OFFSETS:
            print("   {:<5} {}".format(name, self.get_mstatus(name)), end="")
            if i % 6 == 5:
                print()
            i += 1
        print(FMT_NONE)



from abc import ABC, abstractmethod
from typing import Union
from .int32 import Int32


class Immediate:
    """
    This class solves a problem of ambiguity when interpreting assembly code.

    Let's look at the following four cases (assuming 1b is 16 bytes back):
     a) beq     a0, a1, 1b      // conditional jump 16 bytes back
     b) beq     a0, a1, -16     // conditional jump 16 bytes back
     c) addi    a0, a1, 1b      // subtract (pc - 16) from a1
     d) addi    a0, a1, -16     // subtract 16 from a1

    We want a and b to behave the same, but c and d not to.

    The Immediate class solves this problem, by giving each instruction two ways of
    interpreting a given immediate. It can either get the "absolute value" of the
    immediate, or it's relative value to the PC.

    In this case, the beq instruction would interpret both as PC relative values,
    while addi would treat them as absolute values.
    """

    abs_value: Int32
    pcrel_value: Int32

    __slots__ = ["abs_value", "pcrel_value"]

    def __init__(self, abs_value: Union[int, Int32], pcrel_value: Union[int, Int32]):
        self.abs_value = Int32(abs_value)
        self.pcrel_value = Int32(pcrel_value)


class Instruction(ABC):
    name: str
    args: tuple

    @abstractmethod
    def get_imm(self, num: int) -> Immediate:
        """
        parse and get immediate argument
        """
        pass

    @abstractmethod
    def get_reg(self, num: int) -> str:
        """
        parse and get an register argument
        """
        pass

    def __repr__(self):
        return "{} {}".format(self.name, ", ".join(self.args))


class InstructionWithEncoding(ABC):
    """
    Mixin for instructions that have encodings
    """

    @property
    @abstractmethod
    def encoding(self) -> int:
        raise NotImplementedError()



import os
from abc import abstractmethod, ABC
from typing import Union, Iterator, List, ClassVar
from io import IOBase

from . import T_ParserOpts, Program


class ProgramLoader(ABC):
    """
    A program loader is always specific to a given source file. It is a place to store
    all state concerning the parsing and loading of that specific source file, including
    options.
    """

    is_binary: ClassVar[bool]

    def __init__(self, source_name: str, source: IOBase, options: T_ParserOpts):
        self.source_name = source_name
        self.source = source
        self.options = options
        self.filename = os.path.split(self.source_name)[-1]
        self.__post_init__()

    def __post_init__(self):
        pass

    @classmethod
    @abstractmethod
    def can_parse(cls, source_name: str) -> float:
        """
        Return confidence that the file located at source_path
        should be parsed and loaded by this loader
        :param source_name: the name of the input
        :return: the confidence that this file belongs to this parser in [0,1]
        """
        pass

    @classmethod
    @abstractmethod
    def get_options(cls, argv: List[str]) -> [List[str], T_ParserOpts]:
        """
        parse command line args into an options dictionary

        :param argv: the command line args list
        :return: all remaining command line args and the parser options object
        """
        pass

    @classmethod
    def instantiate(
        cls, source_name: str, source: IOBase, options: T_ParserOpts
    ) -> "ProgramLoader":
        """
        Instantiate a loader for the given source file with the required arguments

        :param source_name: the path to the source file
        :param source: IO Object representing the source
        :param options: the parsed options (guaranteed to come from this classes get_options method).
        :return: An instance of a ProgramLoader for the specified source
        """
        return cls(source_name, source, options)

    @abstractmethod
    def parse(self) -> Union[Program, Iterator[Program]]:
        """

        :return:
        """
        pass



"""
RiscEmu (c) 2021 Anton Lydike

SPDX-License-Identifier: MIT
"""

from typing import Tuple, Callable, Dict, Union, Iterable

from abc import ABC

from ..core.exceptions import ASSERT_LEN
from ..core import Instruction, Int32, UInt32, Immediate, CPU, Registers


class InstructionSet(ABC):
    """
    Represents a collection of instructions

    Each instruction set has to inherit from this class. Each instruction should be it's own method:

    instruction_<name>(self, ins: LoadedInstruction)

    instructions containing a dot '.' should replace it with an underscore.
    """

    def __init__(self, cpu: "CPU"):
        """Create a new instance of the Instruction set. This requires access to a CPU, and grabs vertain things
        from it such as access to the MMU and registers.
        """
        self.name = self.__class__.__name__
        self.cpu = cpu

    def load(self) -> Dict[str, Callable[["Instruction"], None]]:
        """
        This is called by the CPU once it instantiates this instruction set

        It returns a dictionary of all instructions in this instruction set,
        pointing to the correct handler for it
        """
        return {name: ins for name, ins in self.get_instructions()}

    def get_instructions(self) -> Iterable[Tuple[str, Callable[[Instruction], None]]]:
        """
        Returns a list of all valid instruction names included in this instruction set

        converts underscores in names to dots
        """
        for member in dir(self):
            if member.startswith("instruction_"):
                yield member[12:].replace("_", "."), getattr(self, member)

    def parse_mem_ins(self, ins: "Instruction") -> Tuple[str, UInt32]:
        """
        parses rd, imm(rs) argument format and returns (rd, imm+rs1)
        (so a register and address tuple for memory instructions)
        """
        assert len(ins.args) == 3
        # handle rd, rs1, imm
        rd = ins.get_reg(0)
        rs = self.regs.get(ins.get_reg(1)).unsigned()
        imm = ins.get_imm(2)
        return rd, rs + imm.abs_value.unsigned()

    def parse_rd_rs_rs(
        self, ins: "Instruction", signed=True
    ) -> Tuple[str, Int32, Int32]:
        """
        Assumes the command is in <name> rd, rs1, rs2 format
        Returns the name of rd, and the values in rs1 and rs2
        """
        ASSERT_LEN(ins.args, 3)
        if signed:
            return (
                ins.get_reg(0),
                Int32(self.regs.get(ins.get_reg(1))),
                Int32(self.regs.get(ins.get_reg(2))),
            )
        else:
            return (
                ins.get_reg(0),
                UInt32(self.regs.get(ins.get_reg(1))),
                UInt32(self.regs.get(ins.get_reg(2))),
            )

    def parse_rd_rs_imm(self, ins: "Instruction") -> Tuple[str, Int32, Immediate]:
        """
        Assumes the command is in <name> rd, rs, imm format
        Returns the name of rd, the value in rs and the immediate imm
        """
        return (
            ins.get_reg(0),
            Int32(self.regs.get(ins.get_reg(1))),
            ins.get_imm(2),
        )

    def parse_rs_rs_imm(self, ins: "Instruction") -> Tuple[Int32, Int32, Immediate]:
        """
        Assumes the command is in <name> rs1, rs2, imm format
        Returns the values in rs1, rs2 and the immediate imm
        """
        return (
            Int32(self.regs.get(ins.get_reg(0))),
            Int32(self.regs.get(ins.get_reg(1))),
            ins.get_imm(2),
        )

    def get_reg_content(self, ins: "Instruction", ind: int) -> Int32:
        """
        get the register name from ins and then return the register contents
        """
        return self.regs.get(ins.get_reg(ind))

    @property
    def pc(self) -> int:
        """
        shorthand for self.cpu.pc
        """
        return self.cpu.pc

    @pc.setter
    def pc(self, val: Union[int, Int32]):
        if isinstance(val, Int32):
            val = val.unsigned_value
        self.cpu.pc = val

    @property
    def mmu(self):
        return self.cpu.mmu

    @property
    def regs(self) -> Registers:
        return self.cpu.regs

    def __repr__(self):
        return "InstructionSet[{}] with {} instructions".format(
            self.__class__.__name__, len(list(self.get_instructions()))
        )
